# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock
import unittest
import zlib
import os

from io import BytesIO
from textwrap import dedent

from itertools import zip_longest
from urllib.parse import quote, parse_qsl
from swift.common import exceptions, internal_client, request_helpers, swob
from swift.common.header_key_dict import HeaderKeyDict
from swift.common.storage_policy import StoragePolicy
from swift.common.middleware.proxy_logging import ProxyLoggingMiddleware
from swift.common.middleware.gatekeeper import GatekeeperMiddleware

from test.debug_logger import debug_logger
from test.unit import with_tempdir, write_fake_ring, patch_policies
from test.unit.common.middleware.helpers import FakeSwift, LeakTrackingIter

from eventlet.green.urllib import request as urllib_request
from eventlet.green.http import client as http_client


class FakeConn(object):
    def __init__(self, body=None):
        if body is None:
            body = []
        self.body = body

    def read(self):
        return json.dumps(self.body).encode('ascii')

    def info(self):
        return {}


def not_sleep(seconds):
    pass


def unicode_string(start, length):
    return u''.join([chr(x) for x in range(start, start + length)])


def path_parts():
    account = unicode_string(1000, 4) + ' ' + unicode_string(1100, 4)
    container = unicode_string(2000, 4) + ' ' + unicode_string(2100, 4)
    obj = unicode_string(3000, 4) + ' ' + unicode_string(3100, 4)
    return account, container, obj


def make_path(account, container=None, obj=None):
    path = '/v1/%s' % quote(account.encode('utf-8'))
    if container:
        path += '/%s' % quote(container.encode('utf-8'))
    if obj:
        path += '/%s' % quote(obj.encode('utf-8'))
    return path


def make_path_info(account, container=None, obj=None):
    # FakeSwift keys on PATH_INFO - which is *encoded* but unquoted
    path = '/v1/%s' % '/'.join(
        p for p in (account, container, obj) if p)
    return swob.bytes_to_wsgi(path.encode('utf-8'))


def get_client_app():
    app = FakeSwift()
    client = internal_client.InternalClient({}, 'test', 1, app=app)
    return client, app


class InternalClient(internal_client.InternalClient):
    def __init__(self):
        pass


class GetMetadataInternalClient(internal_client.InternalClient):
    def __init__(self, test, path, metadata_prefix, acceptable_statuses):
        self.test = test
        self.path = path
        self.metadata_prefix = metadata_prefix
        self.acceptable_statuses = acceptable_statuses
        self.get_metadata_called = 0
        self.metadata = 'some_metadata'

    def _get_metadata(self, path, metadata_prefix, acceptable_statuses=None,
                      headers=None, params=None):
        self.get_metadata_called += 1
        self.test.assertEqual(self.path, path)
        self.test.assertEqual(self.metadata_prefix, metadata_prefix)
        self.test.assertEqual(self.acceptable_statuses, acceptable_statuses)
        return self.metadata


class SetMetadataInternalClient(internal_client.InternalClient):
    def __init__(
            self, test, path, metadata, metadata_prefix, acceptable_statuses):
        self.test = test
        self.path = path
        self.metadata = metadata
        self.metadata_prefix = metadata_prefix
        self.acceptable_statuses = acceptable_statuses
        self.set_metadata_called = 0
        self.metadata = 'some_metadata'

    def _set_metadata(
            self, path, metadata, metadata_prefix='',
            acceptable_statuses=None):
        self.set_metadata_called += 1
        self.test.assertEqual(self.path, path)
        self.test.assertEqual(self.metadata_prefix, metadata_prefix)
        self.test.assertEqual(self.metadata, metadata)
        self.test.assertEqual(self.acceptable_statuses, acceptable_statuses)


class IterInternalClient(internal_client.InternalClient):
    def __init__(
            self, test, path, marker, end_marker, prefix, acceptable_statuses,
            items):
        self.test = test
        self.path = path
        self.marker = marker
        self.end_marker = end_marker
        self.prefix = prefix
        self.acceptable_statuses = acceptable_statuses
        self.items = items

    def _iter_items(
            self, path, marker='', end_marker='', prefix='',
            acceptable_statuses=None):
        self.test.assertEqual(self.path, path)
        self.test.assertEqual(self.marker, marker)
        self.test.assertEqual(self.end_marker, end_marker)
        self.test.assertEqual(self.prefix, prefix)
        self.test.assertEqual(self.acceptable_statuses, acceptable_statuses)
        for item in self.items:
            yield item


class TestCompressingfileReader(unittest.TestCase):
    def test_init(self):
        class CompressObj(object):
            def __init__(self, test, *args):
                self.test = test
                self.args = args

            def method(self, *args):
                self.test.assertEqual(self.args, args)
                return self

        try:
            compressobj = CompressObj(
                self, 9, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)

            old_compressobj = internal_client.compressobj
            internal_client.compressobj = compressobj.method

            f = BytesIO(b'')

            fobj = internal_client.CompressingFileReader(f)
            self.assertEqual(f, fobj._f)
            self.assertEqual(compressobj, fobj._compressor)
            self.assertEqual(False, fobj.done)
            self.assertEqual(True, fobj.first)
            self.assertEqual(0, fobj.crc32)
            self.assertEqual(0, fobj.total_size)
        finally:
            internal_client.compressobj = old_compressobj

    def test_read(self):
        exp_data = b'abcdefghijklmnopqrstuvwxyz'
        fobj = internal_client.CompressingFileReader(
            BytesIO(exp_data), chunk_size=5)

        d = zlib.decompressobj(16 + zlib.MAX_WBITS)
        data = b''.join(d.decompress(chunk)
                        for chunk in iter(fobj.read, b''))

        self.assertEqual(exp_data, data)

    def test_seek(self):
        exp_data = b'abcdefghijklmnopqrstuvwxyz'
        fobj = internal_client.CompressingFileReader(
            BytesIO(exp_data), chunk_size=5)

        # read a couple of chunks only
        for _ in range(2):
            fobj.read()

        # read whole thing after seek and check data
        fobj.seek(0)
        d = zlib.decompressobj(16 + zlib.MAX_WBITS)
        data = b''.join(d.decompress(chunk)
                        for chunk in iter(fobj.read, b''))
        self.assertEqual(exp_data, data)

    def test_seek_not_implemented_exception(self):
        fobj = internal_client.CompressingFileReader(
            BytesIO(b''), chunk_size=5)
        self.assertRaises(NotImplementedError, fobj.seek, 10)
        self.assertRaises(NotImplementedError, fobj.seek, 0, 10)


class TestInternalClient(unittest.TestCase):

    @mock.patch('swift.common.utils.HASH_PATH_SUFFIX', new=b'endcap')
    @with_tempdir
    def test_load_from_config(self, tempdir):
        conf_path = os.path.join(tempdir, 'interal_client.conf')
        conf_body = """
        [DEFAULT]
        swift_dir = %s

        [pipeline:main]
        pipeline = catch_errors cache proxy-server

        [app:proxy-server]
        use = egg:swift#proxy

        [filter:cache]
        use = egg:swift#memcache

        [filter:catch_errors]
        use = egg:swift#catch_errors
        """ % tempdir
        with open(conf_path, 'w') as f:
            f.write(dedent(conf_body))
        account_ring_path = os.path.join(tempdir, 'account.ring.gz')
        write_fake_ring(account_ring_path)
        container_ring_path = os.path.join(tempdir, 'container.ring.gz')
        write_fake_ring(container_ring_path)
        object_ring_path = os.path.join(tempdir, 'object.ring.gz')
        write_fake_ring(object_ring_path)
        logger = debug_logger('test-ic')
        self.assertEqual(logger.get_lines_for_level('warning'), [])
        with patch_policies([StoragePolicy(0, 'legacy', True)]):
            with mock.patch('swift.proxy.server.get_logger',
                            lambda *a, **kw: logger):
                client = internal_client.InternalClient(conf_path, 'test', 1)
            self.assertEqual(client.account_ring,
                             client.app.app.app.account_ring)
            self.assertEqual(client.account_ring.serialized_path,
                             account_ring_path)
            self.assertEqual(client.container_ring,
                             client.app.app.app.container_ring)
            self.assertEqual(client.container_ring.serialized_path,
                             container_ring_path)
            object_ring = client.app.app.app.get_object_ring(0)
            self.assertEqual(client.get_object_ring(0),
                             object_ring)
            self.assertEqual(object_ring.serialized_path,
                             object_ring_path)

    @mock.patch('swift.common.utils.HASH_PATH_SUFFIX', new=b'endcap')
    @with_tempdir
    def test_load_from_config_with_global_conf(self, tempdir):
        account_ring_path = os.path.join(tempdir, 'account.ring.gz')
        write_fake_ring(account_ring_path)
        container_ring_path = os.path.join(tempdir, 'container.ring.gz')
        write_fake_ring(container_ring_path)
        object_ring_path = os.path.join(tempdir, 'object.ring.gz')
        write_fake_ring(object_ring_path)

        # global_conf will override the 'x = y' syntax in conf file...
        conf_path = os.path.join(tempdir, 'internal_client.conf')
        conf_body = """
        [DEFAULT]
        swift_dir = %s
        log_name = conf-file-log-name

        [pipeline:main]
        pipeline = catch_errors cache proxy-server

        [app:proxy-server]
        use = egg:swift#proxy

        [filter:cache]
        use = egg:swift#memcache

        [filter:catch_errors]
        use = egg:swift#catch_errors
        log_name = catch-errors-log-name
        """ % tempdir
        with open(conf_path, 'w') as f:
            f.write(dedent(conf_body))
        global_conf = {'log_name': 'global-conf-log-name'}
        with patch_policies([StoragePolicy(0, 'legacy', True)]):
            client = internal_client.InternalClient(
                conf_path, 'test', 1, global_conf=global_conf)
        self.assertEqual('global-conf-log-name', client.app.logger.server)

        # ...but the 'set x = y' syntax in conf file DEFAULT section will
        # override global_conf
        conf_body = """
        [DEFAULT]
        swift_dir = %s
        set log_name = conf-file-log-name

        [pipeline:main]
        pipeline = catch_errors cache proxy-server

        [app:proxy-server]
        use = egg:swift#proxy

        [filter:cache]
        use = egg:swift#memcache

        [filter:catch_errors]
        use = egg:swift#catch_errors
        log_name = catch-errors-log-name
        """ % tempdir
        with open(conf_path, 'w') as f:
            f.write(dedent(conf_body))
        global_conf = {'log_name': 'global-conf-log-name'}
        with patch_policies([StoragePolicy(0, 'legacy', True)]):
            client = internal_client.InternalClient(
                conf_path, 'test', 1, global_conf=global_conf)
        self.assertEqual('conf-file-log-name', client.app.logger.server)

        # ...and the 'set x = y' syntax in conf file app section will override
        # DEFAULT section and global_conf
        conf_body = """
        [DEFAULT]
        swift_dir = %s
        set log_name = conf-file-log-name

        [pipeline:main]
        pipeline = catch_errors cache proxy-server

        [app:proxy-server]
        use = egg:swift#proxy

        [filter:cache]
        use = egg:swift#memcache

        [filter:catch_errors]
        use = egg:swift#catch_errors
        set log_name = catch-errors-log-name
        """ % tempdir
        with open(conf_path, 'w') as f:
            f.write(dedent(conf_body))
        global_conf = {'log_name': 'global-conf-log-name'}
        with patch_policies([StoragePolicy(0, 'legacy', True)]):
            client = internal_client.InternalClient(
                conf_path, 'test', 1, global_conf=global_conf)
        self.assertEqual('catch-errors-log-name', client.app.logger.server)

    def test_init(self):
        conf_path = 'some_path'
        app = FakeSwift()

        user_agent = 'some_user_agent'
        request_tries = 123

        with mock.patch.object(internal_client, 'loadapp',
                               return_value=app) as mock_loadapp, \
                self.assertRaises(ValueError):
            # First try with a bad arg
            internal_client.InternalClient(
                conf_path, user_agent, request_tries=0)
        mock_loadapp.assert_not_called()

        # if we load it with the gatekeeper middleware then we also get a
        # value error
        gate_keeper_app = GatekeeperMiddleware(app, {})
        gate_keeper_app._pipeline_final_app = app
        gate_keeper_app._pipeline = [gate_keeper_app, app]
        with mock.patch.object(
                internal_client, 'loadapp', return_value=gate_keeper_app) \
                as mock_loadapp, self.assertRaises(ValueError) as err:
            internal_client.InternalClient(
                conf_path, user_agent, request_tries)
        self.assertEqual(
            str(err.exception),
            ('Gatekeeper middleware is not allowed in the InternalClient '
             'proxy pipeline'))

        with mock.patch.object(
                internal_client, 'loadapp', return_value=app) as mock_loadapp:
            client = internal_client.InternalClient(
                conf_path, user_agent, request_tries)

        mock_loadapp.assert_called_once_with(
            conf_path, global_conf=None, allow_modify_pipeline=False)
        self.assertEqual(app, client.app)
        self.assertEqual(user_agent, client.user_agent)
        self.assertEqual(request_tries, client.request_tries)
        self.assertFalse(client.use_replication_network)

        client = internal_client.InternalClient(
            conf_path, user_agent, request_tries, app=app,
            use_replication_network=True)
        self.assertEqual(app, client.app)
        self.assertEqual(user_agent, client.user_agent)
        self.assertEqual(request_tries, client.request_tries)
        self.assertTrue(client.use_replication_network)

        global_conf = {'log_name': 'custom'}
        client = internal_client.InternalClient(
            conf_path, user_agent, request_tries, app=app,
            use_replication_network=True, global_conf=global_conf)
        self.assertEqual(app, client.app)
        self.assertEqual(user_agent, client.user_agent)
        self.assertEqual(request_tries, client.request_tries)
        self.assertTrue(client.use_replication_network)

    def test_init_allow_modify_pipeline(self):
        conf_path = 'some_path'
        app = FakeSwift()
        user_agent = 'some_user_agent'

        with mock.patch.object(internal_client, 'loadapp',
                               return_value=app) as mock_loadapp, \
                self.assertRaises(ValueError) as cm:
            internal_client.InternalClient(
                conf_path, user_agent, 1, allow_modify_pipeline=True)
        mock_loadapp.assert_not_called()
        self.assertIn("'allow_modify_pipeline' is no longer supported",
                      str(cm.exception))

        with mock.patch.object(
                internal_client, 'loadapp', return_value=app) as mock_loadapp:
            internal_client.InternalClient(
                conf_path, user_agent, 1, allow_modify_pipeline=False)
        mock_loadapp.assert_called_once_with(
            conf_path, allow_modify_pipeline=False, global_conf=None)

    def test_gatekeeper_not_loaded(self):
        app = FakeSwift()
        pipeline = [app]

        class RandomMiddleware(object):
            def __init__(self, app):
                self.app = app
                self._pipeline_final_app = app
                self._pipeline = pipeline
                self._pipeline.insert(0, self)

        # if there is no Gatekeeper middleware then it's false
        # just the final app
        self.assertFalse(
            internal_client.InternalClient.check_gatekeeper_not_loaded(app))

        # now with a bunch of middlewares
        app_no_gatekeeper = app
        for i in range(5):
            app_no_gatekeeper = RandomMiddleware(app_no_gatekeeper)
            self.assertFalse(
                internal_client.InternalClient.check_gatekeeper_not_loaded(
                    app_no_gatekeeper))

        # But if we put the gatekeeper on the end, it will be found
        app_with_gatekeeper = GatekeeperMiddleware(app_no_gatekeeper, {})
        pipeline.insert(0, app_with_gatekeeper)
        app_with_gatekeeper._pipeline = pipeline
        with self.assertRaises(ValueError) as err:
            internal_client.InternalClient.check_gatekeeper_not_loaded(
                app_with_gatekeeper)
        self.assertEqual(str(err.exception),
                         ('Gatekeeper middleware is not allowed in the '
                          'InternalClient proxy pipeline'))

        # even if we bury deep into the pipeline
        for i in range(5):
            app_with_gatekeeper = RandomMiddleware(app_with_gatekeeper)
            with self.assertRaises(ValueError) as err:
                internal_client.InternalClient.check_gatekeeper_not_loaded(
                    app_with_gatekeeper)
            self.assertEqual(str(err.exception),
                             ('Gatekeeper middleware is not allowed in the '
                              'InternalClient proxy pipeline'))

    def test_make_request_sets_user_agent(self):
        class FakeApp(FakeSwift):
            def __init__(self, test):
                super(FakeApp, self).__init__()
                self.test = test

            def __call__(self, env, start_response):
                self.test.assertNotIn(
                    'HTTP_X_BACKEND_USE_REPLICATION_NETWORK', env)
                self.test.assertEqual(self.backend_user_agent,
                                      "some_agent")
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 1, use_replication_network=False,
            app=FakeApp(self))
        client.make_request('GET', '/', {}, (200,))

    def test_make_request_clears_txn_id_after_calling_app(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, logger):
                def fake_app(env, start_response):
                    self.app.logger.txn_id = 'foo'
                    self.app.logger.debug('Inside of request')
                    start_response('200 Ok', [('Content-Length', '0')])
                    return []

                self.test = test
                self.user_agent = 'some_agent'
                self.app = fake_app
                self.app.logger = logger
                self.request_tries = 1
                self.use_replication_network = False

        logger = debug_logger()
        # Make sure there's no transaction ID set -- other tests may have
        # polluted the logger
        logger.txn_id = None
        logger.debug('Before request')
        client = InternalClient(self, logger)
        client.make_request('GET', '/', {}, (200,))
        logger.debug('After request')
        self.assertEqual([(args[0], kwargs['extra'].get('txn_id'))
                          for args, kwargs in logger.log_dict['debug']], [
            ('Before request', None),
            ('Inside of request', 'foo'),
            ('After request', None),
        ])

    def test_make_request_defaults_replication_network_header(self):
        class FakeApp(FakeSwift):
            def __init__(self, test):
                super(FakeApp, self).__init__()
                self.test = test
                self.expected_header_value = None

            def __call__(self, env, start_response):
                if self.expected_header_value is None:
                    self.test.assertNotIn(
                        'HTTP_X_BACKEND_USE_REPLICATION_NETWORK', env)
                else:
                    hdr_val = env['HTTP_X_BACKEND_USE_REPLICATION_NETWORK']
                    self.test.assertEqual(self.expected_header_value, hdr_val)
                self.test.assertEqual(self.backend_user_agent,
                                      'some_agent')
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 1, use_replication_network=False,
            app=FakeApp(self))
        client.make_request('GET', '/', {}, (200,))
        # Caller can still override
        client.app.expected_header_value = 'false'
        client.make_request('GET', '/', {
            request_helpers.USE_REPLICATION_NETWORK_HEADER: 'false'}, (200,))
        client.app.expected_header_value = 'true'
        client.make_request('GET', '/', {
            request_helpers.USE_REPLICATION_NETWORK_HEADER: 'true'}, (200,))

        # Switch default behavior
        client.use_replication_network = True

        client.make_request('GET', '/', {}, (200,))
        client.app.expected_header_value = 'false'
        client.make_request('GET', '/', {
            request_helpers.USE_REPLICATION_NETWORK_HEADER: 'false'}, (200,))
        client.app.expected_header_value = 'on'
        client.make_request('GET', '/', {
            request_helpers.USE_REPLICATION_NETWORK_HEADER: 'on'}, (200,))

    def test_make_request_sets_query_string(self):
        captured_envs = []

        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                captured_envs.append(env)
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 1, use_replication_network=False,
            app=FakeApp())
        params = {'param1': 'p1', 'tasty': 'soup'}
        client.make_request('GET', '/', {}, (200,), params=params)
        actual_params = dict(parse_qsl(captured_envs[0]['QUERY_STRING'],
                                       keep_blank_values=True,
                                       strict_parsing=True))
        self.assertEqual(params, actual_params)
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'some_agent')

    def test_make_request_retries(self):
        class FakeApp(FakeSwift):
            def __init__(self):
                super(FakeApp, self).__init__()
                self.request_tries = 4
                self.tries = 0

            def __call__(self, env, start_response):
                self.tries += 1
                if self.tries < self.request_tries:
                    start_response(
                        '500 Internal Server Error', [('Content-Length', '0')])
                else:
                    start_response('200 Ok', [('Content-Length', '0')])
                return []

        class InternalClient(internal_client.InternalClient):
            def __init__(self, *args, **kwargs):
                self.test = kwargs.pop('test')
                super(InternalClient, self).__init__(*args, **kwargs)
                self.sleep_called = 0

            def sleep(self, seconds):
                self.sleep_called += 1
                self.test.assertEqual(2 ** (self.sleep_called), seconds)

        client = InternalClient(
            None, 'some_agent', 4, use_replication_network=False,
            app=FakeApp(), test=self)

        old_sleep = internal_client.sleep
        internal_client.sleep = client.sleep

        try:
            client.make_request('GET', '/', {}, (200,))
        finally:
            internal_client.sleep = old_sleep

        self.assertEqual(3, client.sleep_called)
        self.assertEqual(4, client.app.tries)
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'some_agent')

    def test_base_request_timeout(self):
        # verify that base_request passes timeout arg on to urlopen
        body = {"some": "content"}

        for timeout in (0.0, 42.0, None):
            mocked_func = 'swift.common.internal_client.urllib_request.urlopen'
            with mock.patch(mocked_func) as mock_urlopen:
                mock_urlopen.side_effect = [FakeConn(body)]
                sc = internal_client.SimpleClient('http://0.0.0.0/')
                _, resp_body = sc.base_request('GET', timeout=timeout)
                mock_urlopen.assert_called_once_with(mock.ANY, timeout=timeout)
                # sanity check
                self.assertEqual(body, resp_body)

    def test_base_full_listing(self):
        body1 = [{'name': 'a'}, {'name': "b"}, {'name': "c"}]
        body2 = [{'name': 'd'}]
        body3 = []

        mocked_func = 'swift.common.internal_client.urllib_request.urlopen'
        with mock.patch(mocked_func) as mock_urlopen:
            mock_urlopen.side_effect = [
                FakeConn(body1), FakeConn(body2), FakeConn(body3)]
            sc = internal_client.SimpleClient('http://0.0.0.0/')
            _, resp_body = sc.base_request('GET', full_listing=True)
        self.assertEqual(body1 + body2, resp_body)
        self.assertEqual(3, mock_urlopen.call_count)
        actual_requests = [call[0][0] for call in mock_urlopen.call_args_list]
        self.assertEqual('/?format=json', actual_requests[0].selector)
        self.assertEqual(
            '/?format=json&marker=c', actual_requests[1].selector)
        self.assertEqual(
            '/?format=json&marker=d', actual_requests[2].selector)

    def test_make_request_method_path_headers(self):
        class FakeApp(FakeSwift):
            def __init__(self):
                super(FakeApp, self).__init__()
                self.env = None

            def __call__(self, env, start_response):
                self.env = env
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp())

        for method in 'GET PUT HEAD'.split():
            client.make_request(method, '/', {}, (200,))
            self.assertEqual(client.app.env['REQUEST_METHOD'], method)

        for path in '/one /two/three'.split():
            client.make_request('GET', path, {'X-Test': path}, (200,))
            self.assertEqual(client.app.env['PATH_INFO'], path)
            self.assertEqual(client.app.env['HTTP_X_TEST'], path)
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'some_agent')

    def test_make_request_error_case(self):
        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                body = b'fake error response'
                start_response('409 Conflict',
                               [('Content-Length', str(len(body)))])
                return [body]

        final_fake_app = FakeApp()
        fake_app = ProxyLoggingMiddleware(
            final_fake_app, {}, final_fake_app.logger)
        fake_app._pipeline_final_app = final_fake_app

        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False, app=fake_app)
        with self.assertRaises(internal_client.UnexpectedResponse), \
                mock.patch('swift.common.internal_client.sleep'):
            client.make_request('DELETE', '/container', {}, (200,))

        # Since we didn't provide an X-Timestamp, retrying gives us a chance to
        # succeed (assuming the failure was due to clock skew between servers)
        expected = (' HTTP/1.0 409 ',)
        logger = client.app._pipeline_final_app.logger
        loglines = logger.get_lines_for_level('info')
        for expected, logline in zip_longest(expected, loglines):
            if not expected:
                self.fail('Unexpected extra log line: %r' % logline)
            self.assertIn(expected, logline)
        self.assertEqual(client.app.app.backend_user_agent, 'some_agent')

    def test_make_request_acceptable_status_not_2xx(self):
        class FakeApp(FakeSwift):
            def __init__(self):
                super(FakeApp, self).__init__()
                self.closed_paths = []
                self.fully_read_paths = []
                self.resp_status = None

            def __call__(self, env, start_response):
                body = b'fake error response'
                start_response(self.resp_status,
                               [('Content-Length', str(len(body)))])
                return LeakTrackingIter(body, self.closed_paths.append,
                                        self.fully_read_paths.append,
                                        env['PATH_INFO'])

        def do_test(resp_status):
            final_fake_app = FakeApp()
            fake_app = ProxyLoggingMiddleware(
                final_fake_app, {}, final_fake_app.logger)
            fake_app._pipeline_final_app = final_fake_app
            final_fake_app.resp_status = resp_status
            client = internal_client.InternalClient(
                None, "some_agent", 3, use_replication_network=False,
                app=fake_app)
            expected_error = internal_client.UnexpectedResponse
            with self.assertRaises(expected_error) as ctx, \
                    mock.patch('swift.common.internal_client.sleep'):
                # This is obvious strange tests to expect only 400 Bad Request
                # but this test intended to avoid extra body drain if it's
                # correct object body with 2xx.
                client.make_request('GET', '/cont/obj', {}, (400,))
            logger = client.app._pipeline_final_app.logger
            loglines = logger.get_lines_for_level('info')
            self.assertEqual(client.app.app.backend_user_agent, 'some_agent')
            return (client.app._pipeline_final_app.fully_read_paths,
                    client.app._pipeline_final_app.closed_paths,
                    ctx.exception.resp, loglines)

        fully_read_paths, closed_paths, resp, loglines = do_test('200 OK')
        # Since the 200 is considered "properly handled", it won't be retried
        self.assertEqual(fully_read_paths, [])
        self.assertEqual(closed_paths, [])
        # ...and it'll be on us (the caller) to read and close (for example,
        # by using swob.Response's body property)
        self.assertEqual(resp.body, b'fake error response')
        self.assertEqual(fully_read_paths, ['/cont/obj'])
        self.assertEqual(closed_paths, ['/cont/obj'])

        expected = (' HTTP/1.0 200 ', )
        for expected, logline in zip_longest(expected, loglines):
            if not expected:
                self.fail('Unexpected extra log line: %r' % logline)
            self.assertIn(expected, logline)

        fully_read_paths, closed_paths, resp, loglines = do_test(
            '503 Service Unavailable')
        # But since 5xx is neither "properly handled" not likely to include
        # a large body, it will be retried and responses will already be closed
        self.assertEqual(fully_read_paths, ['/cont/obj'] * 3)
        self.assertEqual(closed_paths, ['/cont/obj'] * 3)

        expected = (' HTTP/1.0 503 ', ' HTTP/1.0 503 ', ' HTTP/1.0 503 ', )
        for expected, logline in zip_longest(expected, loglines):
            if not expected:
                self.fail('Unexpected extra log line: %r' % logline)
            self.assertIn(expected, logline)

    def test_make_request_codes(self):
        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp())

        try:
            old_sleep = internal_client.sleep
            internal_client.sleep = not_sleep

            client.make_request('GET', '/', {}, (200,))
            client.make_request('GET', '/', {}, (2,))
            client.make_request('GET', '/', {}, (400, 200))
            client.make_request('GET', '/', {}, (400, 2))

            with self.assertRaises(internal_client.UnexpectedResponse) \
                    as raised:
                client.make_request('GET', '/', {}, (400,))
            self.assertEqual(200, raised.exception.resp.status_int)

            with self.assertRaises(internal_client.UnexpectedResponse) \
                    as raised:
                client.make_request('GET', '/', {}, (201,))
            self.assertEqual(200, raised.exception.resp.status_int)

            with self.assertRaises(internal_client.UnexpectedResponse) \
                    as raised:
                client.make_request('GET', '/', {}, (111,))
            self.assertTrue(str(raised.exception).startswith(
                'Unexpected response'))
            self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                             'some_agent')
        finally:
            internal_client.sleep = old_sleep

    def test_make_request_calls_fobj_seek_each_try(self):
        class FileObject(object):
            def __init__(self, test):
                self.test = test
                self.seek_called = 0

            def seek(self, offset, whence=0):
                self.seek_called += 1
                self.test.assertEqual(0, offset)
                self.test.assertEqual(0, whence)

        class FakeApp(FakeSwift):
            def __init__(self, status):
                super(FakeApp, self).__init__()
                self.status = status

            def __call__(self, env, start_response):
                start_response(self.status, [('Content-Length', '0')])
                self._calls.append('')
                return []

        def do_test(status, expected_calls):
            fobj = FileObject(self)
            client = internal_client.InternalClient(
                None, 'some_agent', 3, use_replication_network=False,
                app=FakeApp(status))

            with mock.patch.object(internal_client, 'sleep', not_sleep):
                with self.assertRaises(Exception) as exc_mgr:
                    client.make_request('PUT', '/', {}, (2,), fobj)
                self.assertEqual(int(status[:3]),
                                 exc_mgr.exception.resp.status_int)

            self.assertEqual(client.app.call_count, fobj.seek_called)
            self.assertEqual(client.app.call_count, expected_calls)
            self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                             'some_agent')

        do_test('404 Not Found', 1)
        do_test('503 Service Unavailable', 3)

    def test_make_request_request_exception(self):
        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                raise Exception()

        client = internal_client.InternalClient(
            None, 'some_agent', 3, app=FakeApp())
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'some_agent')
        try:
            old_sleep = internal_client.sleep
            internal_client.sleep = not_sleep
            self.assertRaises(
                Exception, client.make_request, 'GET', '/', {}, (2,))
        finally:
            internal_client.sleep = old_sleep

    def test_get_metadata(self):
        class Response(object):
            def __init__(self, headers):
                self.headers = headers
                self.status_int = 200

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, resp_headers):
                self.test = test
                self.path = path
                self.resp_headers = resp_headers
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None, params=None):
                self.make_request_called += 1
                self.test.assertEqual('HEAD', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual((2,), acceptable_statuses)
                self.test.assertIsNone(body_file)
                return Response(self.resp_headers)

        path = 'some_path'
        metadata_prefix = 'some_key-'
        resp_headers = {
            '%sone' % (metadata_prefix): '1',
            '%sTwo' % (metadata_prefix): '2',
            '%sThree' % (metadata_prefix): '3',
            'some_header-four': '4',
            'Some_header-five': '5',
        }
        exp_metadata = {
            'one': '1',
            'two': '2',
            'three': '3',
        }

        client = InternalClient(self, path, resp_headers)
        metadata = client._get_metadata(path, metadata_prefix)
        self.assertEqual(exp_metadata, metadata)
        self.assertEqual(1, client.make_request_called)

    def test_get_metadata_invalid_status(self):

        class FakeApp(FakeSwift):
            def __call__(self, environ, start_response):
                start_response('404 Not Found', [('x-foo', 'bar')])
                return [b'nope']

        client = internal_client.InternalClient(
            None, 'test', 1, use_replication_network=False, app=FakeApp())
        self.assertRaises(internal_client.UnexpectedResponse,
                          client._get_metadata, 'path')
        metadata = client._get_metadata('path', metadata_prefix='x-',
                                        acceptable_statuses=(4,))
        self.assertEqual(metadata, {'foo': 'bar'})
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'test')

    def test_make_path(self):
        account, container, obj = path_parts()
        path = make_path(account, container, obj)

        c = InternalClient()
        self.assertEqual(path, c.make_path(account, container, obj))

    def test_make_path_exception(self):
        c = InternalClient()
        self.assertRaises(ValueError, c.make_path, 'account', None, 'obj')

    def test_iter_items(self):
        class Response(object):
            def __init__(self, status_int, body):
                self.status_int = status_int
                self.body = body

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, responses):
                self.test = test
                self.responses = responses
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.make_request_called += 1
                return self.responses.pop(0)

        exp_items = []
        responses = [Response(200, json.dumps([]).encode('ascii')), ]
        items = []
        client = InternalClient(self, responses)
        for item in client._iter_items('/'):
            items.append(item)
        self.assertEqual(exp_items, items)

        exp_items = []
        responses = []
        for i in range(3):
            data = [
                {'name': 'item%02d' % (2 * i)},
                {'name': 'item%02d' % (2 * i + 1)}]
            responses.append(Response(200, json.dumps(data).encode('ascii')))
            exp_items.extend(data)
        responses.append(Response(204, ''))

        items = []
        client = InternalClient(self, responses)
        for item in client._iter_items('/'):
            items.append(item)
        self.assertEqual(exp_items, items)

    def test_iter_items_with_markers(self):
        class Response(object):
            def __init__(self, status_int, body):
                self.status_int = status_int
                self.body = body.encode('ascii')

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, paths, responses):
                self.test = test
                self.paths = paths
                self.responses = responses

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                exp_path = self.paths.pop(0)
                self.test.assertEqual(exp_path, path)
                return self.responses.pop(0)

        paths = [
            '/?format=json&marker=start&end_marker=end&prefix=',
            '/?format=json&marker=one%C3%A9&end_marker=end&prefix=',
            '/?format=json&marker=two&end_marker=end&prefix=',
        ]

        responses = [
            Response(200, json.dumps([{
                'name': b'one\xc3\xa9'.decode('utf8')}, ])),
            Response(200, json.dumps([{'name': 'two'}, ])),
            Response(204, ''),
        ]

        items = []
        client = InternalClient(self, paths, responses)
        for item in client._iter_items('/', marker='start', end_marker='end'):
            items.append(item['name'].encode('utf8'))

        self.assertEqual(b'one\xc3\xa9 two'.split(), items)

    def test_iter_items_with_markers_and_prefix(self):
        class Response(object):
            def __init__(self, status_int, body):
                self.status_int = status_int
                self.body = body.encode('ascii')

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, paths, responses):
                self.test = test
                self.paths = paths
                self.responses = responses

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                exp_path = self.paths.pop(0)
                self.test.assertEqual(exp_path, path)
                return self.responses.pop(0)

        paths = [
            '/?format=json&marker=prefixed_start&end_marker=prefixed_end'
            '&prefix=prefixed_',
            '/?format=json&marker=prefixed_one%C3%A9&end_marker=prefixed_end'
            '&prefix=prefixed_',
            '/?format=json&marker=prefixed_two&end_marker=prefixed_end'
            '&prefix=prefixed_',
        ]

        responses = [
            Response(200, json.dumps([{
                'name': b'prefixed_one\xc3\xa9'.decode('utf8')}, ])),
            Response(200, json.dumps([{'name': 'prefixed_two'}, ])),
            Response(204, ''),
        ]

        items = []
        client = InternalClient(self, paths, responses)
        for item in client._iter_items('/', marker='prefixed_start',
                                       end_marker='prefixed_end',
                                       prefix='prefixed_'):
            items.append(item['name'].encode('utf8'))

        self.assertEqual(b'prefixed_one\xc3\xa9 prefixed_two'.split(), items)

    def test_iter_item_read_response_if_status_is_acceptable(self):
        class Response(object):
            def __init__(self, status_int, body, app_iter):
                self.status_int = status_int
                self.body = body
                self.app_iter = app_iter

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, responses):
                self.test = test
                self.responses = responses

            def make_request(
                self, method, path, headers, acceptable_statuses,
                    body_file=None):
                resp = self.responses.pop(0)
                if resp.status_int in acceptable_statuses or \
                        resp.status_int // 100 in acceptable_statuses:
                    return resp
                if resp:
                    raise internal_client.UnexpectedResponse(
                        'Unexpected response: %s' % resp.status_int, resp)

        num_list = []

        def generate_resp_body():
            for i in range(1, 5):
                yield str(i).encode('ascii')
                num_list.append(i)

        exp_items = []
        responses = [Response(204, json.dumps([]).encode('ascii'),
                              generate_resp_body())]
        items = []
        client = InternalClient(self, responses)
        for item in client._iter_items('/'):
            items.append(item)
        self.assertEqual(exp_items, items)
        self.assertEqual(len(num_list), 0)

        responses = [Response(300, json.dumps([]).encode('ascii'),
                              generate_resp_body())]
        client = InternalClient(self, responses)
        self.assertRaises(internal_client.UnexpectedResponse,
                          next, client._iter_items('/'))

        exp_items = []
        responses = [Response(404, json.dumps([]).encode('ascii'),
                              generate_resp_body())]
        items = []
        client = InternalClient(self, responses)
        for item in client._iter_items('/'):
            items.append(item)
        self.assertEqual(exp_items, items)
        self.assertEqual(len(num_list), 4)

    def test_set_metadata(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, exp_headers):
                self.test = test
                self.path = path
                self.exp_headers = exp_headers
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.make_request_called += 1
                self.test.assertEqual('POST', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual(self.exp_headers, headers)
                self.test.assertEqual((2,), acceptable_statuses)
                self.test.assertIsNone(body_file)

        path = 'some_path'
        metadata_prefix = 'some_key-'
        metadata = {
            '%sone' % (metadata_prefix): '1',
            '%stwo' % (metadata_prefix): '2',
            'three': '3',
        }
        exp_headers = {
            '%sone' % (metadata_prefix): '1',
            '%stwo' % (metadata_prefix): '2',
            '%sthree' % (metadata_prefix): '3',
        }

        client = InternalClient(self, path, exp_headers)
        client._set_metadata(path, metadata, metadata_prefix)
        self.assertEqual(1, client.make_request_called)

    def test_iter_containers(self):
        account, container, obj = path_parts()
        path = make_path(account)
        items = '0 1 2'.split()
        marker = 'some_marker'
        end_marker = 'some_end_marker'
        prefix = 'some_prefix'
        acceptable_statuses = 'some_status_list'
        client = IterInternalClient(
            self, path, marker, end_marker, prefix, acceptable_statuses, items)
        ret_items = []
        for container in client.iter_containers(
                account, marker, end_marker, prefix,
                acceptable_statuses=acceptable_statuses):
            ret_items.append(container)
        self.assertEqual(items, ret_items)

    def test_create_account(self):
        account, container, obj = path_parts()
        path = make_path_info(account)
        client, app = get_client_app()
        app.register('PUT', path, swob.HTTPCreated, {})
        client.create_account(account)
        self.assertEqual([('PUT', path, {
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual(app.backend_user_agent, 'test')
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)

    def test_delete_account(self):
        account, container, obj = path_parts()
        path = make_path_info(account)
        client, app = get_client_app()
        app.register('DELETE', path, swob.HTTPNoContent, {})
        client.delete_account(account)
        self.assertEqual(1, len(app._calls))
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_get_account_info(self):
        class Response(object):
            def __init__(self, containers, objects):
                self.headers = {
                    'x-account-container-count': containers,
                    'x-account-object-count': objects,
                }
                self.status_int = 200

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, resp):
                self.test = test
                self.path = path
                self.resp = resp

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.test.assertEqual('HEAD', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual({}, headers)
                self.test.assertEqual((2, 404), acceptable_statuses)
                self.test.assertIsNone(body_file)
                return self.resp

        account, container, obj = path_parts()
        path = make_path(account)
        containers, objects = 10, 100
        client = InternalClient(self, path, Response(containers, objects))
        info = client.get_account_info(account)
        self.assertEqual((containers, objects), info)

    def test_get_account_info_404(self):
        class Response(object):
            def __init__(self):
                self.headers = {
                    'x-account-container-count': 10,
                    'x-account-object-count': 100,
                }
                self.status_int = 404

        class InternalClient(internal_client.InternalClient):
            def __init__(self):
                pass

            def make_path(self, *a, **kw):
                return 'some_path'

            def make_request(self, *a, **kw):
                return Response()

        client = InternalClient()
        info = client.get_account_info('some_account')
        self.assertEqual((0, 0), info)

    def test_get_account_metadata(self):
        account, container, obj = path_parts()
        path = make_path(account)
        acceptable_statuses = 'some_status_list'
        metadata_prefix = 'some_metadata_prefix'
        client = GetMetadataInternalClient(
            self, path, metadata_prefix, acceptable_statuses)
        metadata = client.get_account_metadata(
            account, metadata_prefix, acceptable_statuses)
        self.assertEqual(client.metadata, metadata)
        self.assertEqual(1, client.get_metadata_called)

    def test_get_metadadata_with_acceptable_status(self):
        account, container, obj = path_parts()
        path = make_path_info(account)
        client, app = get_client_app()
        resp_headers = {'some-important-header': 'some value'}
        app.register('GET', path, swob.HTTPOk, resp_headers)
        metadata = client.get_account_metadata(
            account, acceptable_statuses=(2, 4))
        self.assertEqual(metadata['some-important-header'],
                         'some value')
        app.register('GET', path, swob.HTTPNotFound, resp_headers)
        metadata = client.get_account_metadata(
            account, acceptable_statuses=(2, 4))
        self.assertEqual(metadata['some-important-header'],
                         'some value')
        app.register('GET', path, swob.HTTPServerError, resp_headers)
        self.assertRaises(internal_client.UnexpectedResponse,
                          client.get_account_metadata, account,
                          acceptable_statuses=(2, 4))

    def test_set_account_metadata(self):
        account, container, obj = path_parts()
        path = make_path_info(account)
        client, app = get_client_app()
        app.register('POST', path, swob.HTTPAccepted, {})
        client.set_account_metadata(account, {'Color': 'Blue'},
                                    metadata_prefix='X-Account-Meta-')
        self.assertEqual([('POST', path, {
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'X-Account-Meta-Color': 'Blue',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_set_account_metadata_plumbing(self):
        account, container, obj = path_parts()
        path = make_path(account)
        metadata = 'some_metadata'
        metadata_prefix = 'some_metadata_prefix'
        acceptable_statuses = 'some_status_list'
        client = SetMetadataInternalClient(
            self, path, metadata, metadata_prefix, acceptable_statuses)
        client.set_account_metadata(
            account, metadata, metadata_prefix, acceptable_statuses)
        self.assertEqual(1, client.set_metadata_called)

    def test_container_exists(self):
        class Response(object):
            def __init__(self, status_int):
                self.status_int = status_int

        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, resp):
                self.test = test
                self.path = path
                self.make_request_called = 0
                self.resp = resp

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.make_request_called += 1
                self.test.assertEqual('HEAD', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual({}, headers)
                self.test.assertEqual((2, 404), acceptable_statuses)
                self.test.assertIsNone(body_file)
                return self.resp

        account, container, obj = path_parts()
        path = make_path(account, container)

        client = InternalClient(self, path, Response(200))
        self.assertEqual(True, client.container_exists(account, container))
        self.assertEqual(1, client.make_request_called)

        client = InternalClient(self, path, Response(404))
        self.assertEqual(False, client.container_exists(account, container))
        self.assertEqual(1, client.make_request_called)

    def test_create_container(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container)
        client, app = get_client_app()
        app.register('PUT', path, swob.HTTPCreated, {})
        client.create_container(account, container)
        self.assertEqual([('PUT', path, {
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual(app.backend_user_agent, 'test')
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)

    def test_create_container_plumbing(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, headers):
                self.test = test
                self.path = path
                self.headers = headers
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.make_request_called += 1
                self.test.assertEqual('PUT', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual(self.headers, headers)
                self.test.assertEqual((2,), acceptable_statuses)
                self.test.assertIsNone(body_file)

        account, container, obj = path_parts()
        path = make_path(account, container)
        headers = 'some_headers'
        client = InternalClient(self, path, headers)
        client.create_container(account, container, headers)
        self.assertEqual(1, client.make_request_called)

    def test_delete_container(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container)
        client, app = get_client_app()
        app.register('DELETE', path, swob.HTTPNoContent, {})
        client.delete_container(account, container)
        self.assertEqual(1, len(app._calls))
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_delete_container_plumbing(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path):
                self.test = test
                self.path = path
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None):
                self.make_request_called += 1
                self.test.assertEqual('DELETE', method)
                self.test.assertEqual(self.path, path)
                self.test.assertEqual({}, headers)
                self.test.assertEqual((2, 404), acceptable_statuses)
                self.test.assertIsNone(body_file)

        account, container, obj = path_parts()
        path = make_path(account, container)
        client = InternalClient(self, path)
        client.delete_container(account, container)
        self.assertEqual(1, client.make_request_called)

    def test_get_container_metadata(self):
        account, container, obj = path_parts()
        path = make_path(account, container)
        metadata_prefix = 'some_metadata_prefix'
        acceptable_statuses = 'some_status_list'
        client = GetMetadataInternalClient(
            self, path, metadata_prefix, acceptable_statuses)
        metadata = client.get_container_metadata(
            account, container, metadata_prefix, acceptable_statuses)
        self.assertEqual(client.metadata, metadata)
        self.assertEqual(1, client.get_metadata_called)

    def test_iter_objects(self):
        account, container, obj = path_parts()
        path = make_path(account, container)
        marker = 'some_maker'
        end_marker = 'some_end_marker'
        prefix = 'some_prefix'
        acceptable_statuses = 'some_status_list'
        items = '0 1 2'.split()
        client = IterInternalClient(
            self, path, marker, end_marker, prefix, acceptable_statuses, items)
        ret_items = []
        for obj in client.iter_objects(
                account, container, marker, end_marker, prefix,
                acceptable_statuses):
            ret_items.append(obj)
        self.assertEqual(items, ret_items)

    def test_set_container_metadata(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container)
        client, app = get_client_app()
        app.register('POST', path, swob.HTTPAccepted, {})
        client.set_container_metadata(account, container, {'Color': 'Blue'},
                                      metadata_prefix='X-Container-Meta-')
        self.assertEqual([('POST', path, {
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'X-Container-Meta-Color': 'Blue',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_set_container_metadata_plumbing(self):
        account, container, obj = path_parts()
        path = make_path(account, container)
        metadata = 'some_metadata'
        metadata_prefix = 'some_metadata_prefix'
        acceptable_statuses = 'some_status_list'
        client = SetMetadataInternalClient(
            self, path, metadata, metadata_prefix, acceptable_statuses)
        client.set_container_metadata(
            account, container, metadata, metadata_prefix, acceptable_statuses)
        self.assertEqual(1, client.set_metadata_called)

    def test_delete_object(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container, obj)
        client, app = get_client_app()
        app.register('DELETE', path, swob.HTTPNoContent, {})
        client.delete_object(account, container, obj)
        self.assertEqual(app.unclosed_requests, {})
        self.assertEqual(1, len(app._calls))
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

        app.register('DELETE', path, swob.HTTPNotFound, {})
        client.delete_object(account, container, obj)
        self.assertEqual(app.unclosed_requests, {})
        self.assertEqual(2, len(app._calls))
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)

    def test_get_object_metadata(self):
        account, container, obj = path_parts()
        path = make_path(account, container, obj)
        metadata_prefix = 'some_metadata_prefix'
        acceptable_statuses = 'some_status_list'
        client = GetMetadataInternalClient(
            self, path, metadata_prefix, acceptable_statuses)
        metadata = client.get_object_metadata(
            account, container, obj, metadata_prefix,
            acceptable_statuses)
        self.assertEqual(client.metadata, metadata)
        self.assertEqual(1, client.get_metadata_called)

    def test_get_metadata_extra_headers(self):
        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                self.req_env = env
                start_response('200 Ok', [('Content-Length', '0')])
                return []

        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp())
        headers = {'X-Foo': 'bar'}
        client.get_object_metadata('account', 'container', 'obj',
                                   headers=headers)
        self.assertEqual(client.app.req_env['HTTP_X_FOO'], 'bar')

    def test_get_object(self):
        account, container, obj = path_parts()
        path_info = make_path_info(account, container, obj)
        client, app = get_client_app()
        headers = {'foo': 'bar'}
        body = b'some_object_body'
        params = {'symlink': 'get'}
        app.register('GET', path_info, swob.HTTPOk, headers, body)
        req_headers = {'x-important-header': 'some_important_value'}
        status_int, resp_headers, obj_iter = client.get_object(
            account, container, obj, req_headers, params=params)
        self.assertEqual(status_int // 100, 2)
        for k, v in headers.items():
            self.assertEqual(v, resp_headers[k])
        self.assertEqual(b''.join(obj_iter), body)
        self.assertEqual(resp_headers['content-length'], str(len(body)))
        self.assertEqual(app.call_count, 1)
        req_headers.update({
            'host': 'localhost:80',  # from swob.Request.blank
            'user-agent': 'test',  # from IC
            'x-backend-allow-reserved-names': 'true',  # also from IC
        })
        self.assertEqual(app.calls_with_headers, [(
            'GET', path_info + '?symlink=get', HeaderKeyDict(req_headers))])

    def test_iter_object_lines(self):
        class FakeApp(FakeSwift):
            def __init__(self, lines):
                super(FakeApp, self).__init__()
                self.lines = lines

            def __call__(self, env, start_response):
                start_response('200 Ok', [('Content-Length', '0')])
                return [b'%s\n' % x for x in self.lines]

        lines = b'line1 line2 line3'.split()
        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp(lines))
        ret_lines = []
        for line in client.iter_object_lines('account', 'container', 'object'):
            ret_lines.append(line)
        self.assertEqual(lines, ret_lines)
        self.assertEqual(client.app._pipeline_final_app.backend_user_agent,
                         'some_agent')

    def test_iter_object_lines_compressed_object(self):
        class FakeApp(FakeSwift):
            def __init__(self, lines):
                super(FakeApp, self).__init__()
                self.lines = lines

            def __call__(self, env, start_response):
                start_response('200 Ok', [('Content-Length', '0')])
                return internal_client.CompressingFileReader(
                    BytesIO(b'\n'.join(self.lines)))

        lines = b'line1 line2 line3'.split()
        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp(lines))
        ret_lines = []
        for line in client.iter_object_lines(
                'account', 'container', 'object.gz'):
            ret_lines.append(line)
        self.assertEqual(lines, ret_lines)

    def test_iter_object_lines_404(self):
        class FakeApp(FakeSwift):
            def __call__(self, env, start_response):
                start_response('404 Not Found', [])
                return [b'one\ntwo\nthree']

        client = internal_client.InternalClient(
            None, 'some_agent', 3, use_replication_network=False,
            app=FakeApp())
        lines = []
        for line in client.iter_object_lines(
                'some_account', 'some_container', 'some_object',
                acceptable_statuses=(2, 404)):
            lines.append(line)
        self.assertEqual([], lines)

    def test_set_object_metadata(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container, obj)
        client, app = get_client_app()
        app.register('POST', path, swob.HTTPAccepted, {})
        client.set_object_metadata(account, container, obj, {'Color': 'Blue'},
                                   metadata_prefix='X-Object-Meta-')
        self.assertEqual([('POST', path, {
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'X-Object-Meta-Color': 'Blue',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_set_object_metadata_plumbing(self):
        account, container, obj = path_parts()
        path = make_path(account, container, obj)
        metadata = 'some_metadata'
        metadata_prefix = 'some_metadata_prefix'
        acceptable_statuses = 'some_status_list'
        client = SetMetadataInternalClient(
            self, path, metadata, metadata_prefix, acceptable_statuses)
        client.set_object_metadata(
            account, container, obj, metadata, metadata_prefix,
            acceptable_statuses)
        self.assertEqual(1, client.set_metadata_called)

    def test_upload_object(self):
        account, container, obj = path_parts()
        path = make_path_info(account, container, obj)
        client, app = get_client_app()
        app.register('PUT', path, swob.HTTPCreated, {})
        client.upload_object(BytesIO(b'fobj'), account, container, obj)
        self.assertEqual([('PUT', path, {
            'Transfer-Encoding': 'chunked',
            'X-Backend-Allow-Reserved-Names': 'true',
            'Host': 'localhost:80',
            'User-Agent': 'test'
        })], app.calls_with_headers)
        self.assertEqual({}, app.unread_requests)
        self.assertEqual({}, app.unclosed_requests)
        self.assertEqual(app.backend_user_agent, 'test')

    def test_upload_object_plumbing(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, headers, fobj):
                self.test = test
                self.use_replication_network = False
                self.path = path
                self.headers = headers
                self.fobj = fobj
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None, params=None):
                self.make_request_called += 1
                self.test.assertEqual(self.path, path)
                exp_headers = dict(self.headers)
                exp_headers['Transfer-Encoding'] = 'chunked'
                self.test.assertEqual(exp_headers, headers)
                self.test.assertEqual(self.fobj, fobj)

        fobj = 'some_fobj'
        account, container, obj = path_parts()
        path = make_path(account, container, obj)
        headers = {'key': 'value'}

        client = InternalClient(self, path, headers, fobj)
        client.upload_object(fobj, account, container, obj, headers)
        self.assertEqual(1, client.make_request_called)

    def test_upload_object_not_chunked(self):
        class InternalClient(internal_client.InternalClient):
            def __init__(self, test, path, headers, fobj):
                self.test = test
                self.path = path
                self.headers = headers
                self.fobj = fobj
                self.make_request_called = 0

            def make_request(
                    self, method, path, headers, acceptable_statuses,
                    body_file=None, params=None):
                self.make_request_called += 1
                self.test.assertEqual(self.path, path)
                exp_headers = dict(self.headers)
                self.test.assertEqual(exp_headers, headers)
                self.test.assertEqual(self.fobj, fobj)

        fobj = 'some_fobj'
        account, container, obj = path_parts()
        path = make_path(account, container, obj)
        headers = {'key': 'value', 'Content-Length': len(fobj)}

        client = InternalClient(self, path, headers, fobj)
        client.upload_object(fobj, account, container, obj, headers)
        self.assertEqual(1, client.make_request_called)


class TestGetAuth(unittest.TestCase):
    @mock.patch.object(urllib_request, 'urlopen')
    @mock.patch.object(urllib_request, 'Request')
    def test_ok(self, request, urlopen):
        def getheader(name):
            d = {'X-Storage-Url': 'url', 'X-Auth-Token': 'token'}
            return d.get(name)
        urlopen.return_value.info.return_value.getheader = getheader

        url, token = internal_client.get_auth(
            'http://127.0.0.1', 'user', 'key')

        self.assertEqual(url, "url")
        self.assertEqual(token, "token")
        request.assert_called_with('http://127.0.0.1')
        request.return_value.add_header.assert_any_call('X-Auth-User', 'user')
        request.return_value.add_header.assert_any_call('X-Auth-Key', 'key')

    def test_invalid_version(self):
        self.assertRaises(SystemExit, internal_client.get_auth,
                          'http://127.0.0.1', 'user', 'key', auth_version=2.0)


class TestSimpleClient(unittest.TestCase):

    def _test_get_head(self, request, urlopen, method):

        mock_time_value = [1401224049.98]

        def mock_time():
            # global mock_time_value
            mock_time_value[0] += 1
            return mock_time_value[0]

        with mock.patch('swift.common.internal_client.time', mock_time):
            # basic request, only url as kwarg
            request.return_value.get_type.return_value = "http"
            urlopen.return_value.read.return_value = b''
            urlopen.return_value.getcode.return_value = 200
            urlopen.return_value.info.return_value = {'content-length': '345'}
            sc = internal_client.SimpleClient(url='http://127.0.0.1')
            logger = debug_logger('test-ic')
            retval = sc.retry_request(
                method, headers={'content-length': '123'}, logger=logger)
            self.assertEqual(urlopen.call_count, 1)
            request.assert_called_with('http://127.0.0.1?format=json',
                                       headers={'content-length': '123'},
                                       data=None)
            self.assertEqual([{'content-length': '345'}, None], retval)
            self.assertEqual(method, request.return_value.get_method())
            self.assertEqual(logger.get_lines_for_level('debug'), [
                '-> 2014-05-27T20:54:11 ' + method +
                ' http://127.0.0.1%3Fformat%3Djson 200 '
                '123 345 1401224050.98 1401224051.98 1.0 -'
            ])

            # Check if JSON is decoded
            urlopen.return_value.read.return_value = b'{}'
            retval = sc.retry_request(method)
            self.assertEqual([{'content-length': '345'}, {}], retval)

            # same as above, now with token
            sc = internal_client.SimpleClient(url='http://127.0.0.1',
                                              token='token')
            retval = sc.retry_request(method)
            request.assert_called_with('http://127.0.0.1?format=json',
                                       headers={'X-Auth-Token': 'token'},
                                       data=None)
            self.assertEqual([{'content-length': '345'}, {}], retval)

            # same as above, now with prefix
            sc = internal_client.SimpleClient(url='http://127.0.0.1',
                                              token='token')
            retval = sc.retry_request(method, prefix="pre_")
            request.assert_called_with(
                'http://127.0.0.1?format=json&prefix=pre_',
                headers={'X-Auth-Token': 'token'}, data=None)
            self.assertEqual([{'content-length': '345'}, {}], retval)

            # same as above, now with container name
            retval = sc.retry_request(method, container='cont')
            request.assert_called_with('http://127.0.0.1/cont?format=json',
                                       headers={'X-Auth-Token': 'token'},
                                       data=None)
            self.assertEqual([{'content-length': '345'}, {}], retval)

            # same as above, now with object name
            retval = sc.retry_request(method, container='cont', name='obj')
            request.assert_called_with('http://127.0.0.1/cont/obj',
                                       headers={'X-Auth-Token': 'token'},
                                       data=None)
            self.assertEqual([{'content-length': '345'}, {}], retval)

    @mock.patch.object(urllib_request, 'urlopen')
    @mock.patch.object(urllib_request, 'Request')
    def test_get(self, request, urlopen):
        self._test_get_head(request, urlopen, 'GET')

    @mock.patch.object(urllib_request, 'urlopen')
    @mock.patch.object(urllib_request, 'Request')
    def test_head(self, request, urlopen):
        self._test_get_head(request, urlopen, 'HEAD')

    @mock.patch.object(urllib_request, 'urlopen')
    @mock.patch.object(urllib_request, 'Request')
    def test_get_with_retries_all_failed(self, request, urlopen):
        # Simulate a failing request, ensure retries done
        request.return_value.get_type.return_value = "http"
        urlopen.side_effect = urllib_request.URLError('')
        sc = internal_client.SimpleClient(url='http://127.0.0.1', retries=1)
        with mock.patch('swift.common.internal_client.sleep') as mock_sleep:
            self.assertRaises(urllib_request.URLError, sc.retry_request, 'GET')
        self.assertEqual(mock_sleep.call_count, 1)
        self.assertEqual(request.call_count, 2)
        self.assertEqual(urlopen.call_count, 2)

    @mock.patch.object(urllib_request, 'urlopen')
    @mock.patch.object(urllib_request, 'Request')
    def test_get_with_retries(self, request, urlopen):
        # First request fails, retry successful
        request.return_value.get_type.return_value = "http"
        mock_resp = mock.MagicMock()
        mock_resp.read.return_value = b''
        mock_resp.info.return_value = {}
        urlopen.side_effect = [urllib_request.URLError(''), mock_resp]
        sc = internal_client.SimpleClient(url='http://127.0.0.1', retries=1,
                                          token='token')

        with mock.patch('swift.common.internal_client.sleep') as mock_sleep:
            retval = sc.retry_request('GET')
        self.assertEqual(mock_sleep.call_count, 1)
        self.assertEqual(request.call_count, 2)
        self.assertEqual(urlopen.call_count, 2)
        request.assert_called_with('http://127.0.0.1?format=json', data=None,
                                   headers={'X-Auth-Token': 'token'})
        self.assertEqual([{}, None], retval)
        self.assertEqual(sc.attempts, 2)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_get_with_retries_param(self, mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        mock_response.info.return_value = {}
        mock_urlopen.side_effect = http_client.BadStatusLine('')
        c = internal_client.SimpleClient(url='http://127.0.0.1', token='token')
        self.assertEqual(c.retries, 5)

        # first without retries param
        with mock.patch('swift.common.internal_client.sleep') as mock_sleep:
            self.assertRaises(http_client.BadStatusLine,
                              c.retry_request, 'GET')
        self.assertEqual(mock_sleep.call_count, 5)
        self.assertEqual(mock_urlopen.call_count, 6)
        # then with retries param
        mock_urlopen.reset_mock()
        with mock.patch('swift.common.internal_client.sleep') as mock_sleep:
            self.assertRaises(http_client.BadStatusLine,
                              c.retry_request, 'GET', retries=2)
        self.assertEqual(mock_sleep.call_count, 2)
        self.assertEqual(mock_urlopen.call_count, 3)
        # and this time with a real response
        mock_urlopen.reset_mock()
        mock_urlopen.side_effect = [http_client.BadStatusLine(''),
                                    mock_response]
        with mock.patch('swift.common.internal_client.sleep') as mock_sleep:
            retval = c.retry_request('GET', retries=1)
        self.assertEqual(mock_sleep.call_count, 1)
        self.assertEqual(mock_urlopen.call_count, 2)
        self.assertEqual([{}, None], retval)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_request_with_retries_with_HTTPError(self, mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        c = internal_client.SimpleClient(url='http://127.0.0.1', token='token')
        self.assertEqual(c.retries, 5)

        for request_method in 'GET PUT POST DELETE HEAD COPY'.split():
            mock_urlopen.reset_mock()
            mock_urlopen.side_effect = urllib_request.HTTPError(*[None] * 5)
            with mock.patch('swift.common.internal_client.sleep') \
                    as mock_sleep:
                self.assertRaises(exceptions.ClientException,
                                  c.retry_request, request_method, retries=1)
            self.assertEqual(mock_sleep.call_count, 1)
            self.assertEqual(mock_urlopen.call_count, 2)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_request_container_with_retries_with_HTTPError(self,
                                                           mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        c = internal_client.SimpleClient(url='http://127.0.0.1', token='token')
        self.assertEqual(c.retries, 5)

        for request_method in 'GET PUT POST DELETE HEAD COPY'.split():
            mock_urlopen.reset_mock()
            mock_urlopen.side_effect = urllib_request.HTTPError(*[None] * 5)
            with mock.patch('swift.common.internal_client.sleep') \
                    as mock_sleep:
                self.assertRaises(exceptions.ClientException,
                                  c.retry_request, request_method,
                                  container='con', retries=1)
            self.assertEqual(mock_sleep.call_count, 1)
            self.assertEqual(mock_urlopen.call_count, 2)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_request_object_with_retries_with_HTTPError(self,
                                                        mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        c = internal_client.SimpleClient(url='http://127.0.0.1', token='token')
        self.assertEqual(c.retries, 5)

        for request_method in 'GET PUT POST DELETE HEAD COPY'.split():
            mock_urlopen.reset_mock()
            mock_urlopen.side_effect = urllib_request.HTTPError(*[None] * 5)
            with mock.patch('swift.common.internal_client.sleep') \
                    as mock_sleep:
                self.assertRaises(exceptions.ClientException,
                                  c.retry_request, request_method,
                                  container='con', name='obj', retries=1)
            self.assertEqual(mock_sleep.call_count, 1)
            self.assertEqual(mock_urlopen.call_count, 2)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_delete_object_with_404_no_retry(self, mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        err_args = [None, 404, None, None, None]
        mock_urlopen.side_effect = urllib_request.HTTPError(*err_args)

        with mock.patch('swift.common.internal_client.sleep') as mock_sleep, \
                self.assertRaises(exceptions.ClientException) as caught:
            internal_client.delete_object('http://127.0.0.1',
                                          container='con', name='obj')
        self.assertEqual(caught.exception.http_status, 404)
        self.assertEqual(mock_sleep.call_count, 0)
        self.assertEqual(mock_urlopen.call_count, 1)

    @mock.patch.object(urllib_request, 'urlopen')
    def test_delete_object_with_409_no_retry(self, mock_urlopen):
        mock_response = mock.MagicMock()
        mock_response.read.return_value = b''
        err_args = [None, 409, None, None, None]
        mock_urlopen.side_effect = urllib_request.HTTPError(*err_args)

        with mock.patch('swift.common.internal_client.sleep') as mock_sleep, \
                self.assertRaises(exceptions.ClientException) as caught:
            internal_client.delete_object('http://127.0.0.1',
                                          container='con', name='obj')
        self.assertEqual(caught.exception.http_status, 409)
        self.assertEqual(mock_sleep.call_count, 0)
        self.assertEqual(mock_urlopen.call_count, 1)

    def test_proxy(self):
        # check that proxy arg is passed through to the urllib Request
        scheme = 'http'
        proxy_host = '127.0.0.1:80'
        proxy = '%s://%s' % (scheme, proxy_host)
        url = 'https://127.0.0.1:1/a'

        mocked = 'swift.common.internal_client.urllib_request.urlopen'

        # module level methods
        for func in (internal_client.put_object,
                     internal_client.delete_object):
            with mock.patch(mocked) as mock_urlopen:
                mock_urlopen.return_value = FakeConn()
                func(url, container='c', name='o1', contents='', proxy=proxy,
                     timeout=0.1, retries=0)
                self.assertEqual(1, mock_urlopen.call_count)
                args, kwargs = mock_urlopen.call_args
                self.assertEqual(1, len(args))
                self.assertEqual(1, len(kwargs))
                self.assertEqual(0.1, kwargs['timeout'])
                self.assertIsInstance(args[0], urllib_request.Request)
                self.assertEqual(proxy_host, args[0].host)
                # TODO: figure out why this doesn't match `scheme`, whether
                # py2 (where it did) or py3 is messed up, whether we care,
                # and what can be done about it
                self.assertEqual('https', args[0].type)

        # class methods
        content = mock.MagicMock()
        cl = internal_client.SimpleClient(url)
        scenarios = ((cl.get_account, []),
                     (cl.get_container, ['c']),
                     (cl.put_container, ['c']),
                     (cl.put_object, ['c', 'o', content]))
        for scenario in scenarios:
            with mock.patch(mocked) as mock_urlopen:
                mock_urlopen.return_value = FakeConn()
                scenario[0](*scenario[1], proxy=proxy, timeout=0.1)
                self.assertEqual(1, mock_urlopen.call_count)
                args, kwargs = mock_urlopen.call_args
                self.assertEqual(1, len(args))
                self.assertEqual(1, len(kwargs))
                self.assertEqual(0.1, kwargs['timeout'])
                self.assertIsInstance(args[0], urllib_request.Request)
                self.assertEqual(proxy_host, args[0].host)
                # See above
                self.assertEqual('https', args[0].type)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import errno
import hashlib
import logging
from logging.handlers import SysLogHandler
import os
import socket
import stat
import string
import sys
import time
import fcntl
import eventlet
import datetime

from swift.common.utils.base import md5, quote, split_path
from swift.common.utils.timestamp import UTC
from swift.common.utils.config import config_true_value
# common.utils imports a fully qualified common.exceptions so that
# common.exceptions can import common.utils with out a circular import error
# (if we only make reference to attributes of a module w/i our function/method
# bodies fully qualifed module names can have their attributes lazily
# evaluated); as the only other module with-in utils that imports exceptions:
# we do the same here
import swift.common.exceptions

from eventlet.green.http import client as green_http_client
import http.client
from eventlet.green import threading


NOTICE = 25

LOG_LINE_DEFAULT_FORMAT = '{remote_addr} - - [{time.d}/{time.b}/{time.Y}' \
                          ':{time.H}:{time.M}:{time.S} +0000] ' \
                          '"{method} {path}" {status} {content_length} ' \
                          '"{referer}" "{txn_id}" "{user_agent}" ' \
                          '{trans_time:.4f} "{additional_info}" {pid} ' \
                          '{policy_index}'


def logging_monkey_patch():
    # explicitly patch the logging lock
    logging._lock = logging.threading.RLock()
    # setup notice level logging
    logging.addLevelName(NOTICE, 'NOTICE')
    SysLogHandler.priority_map['NOTICE'] = 'notice'
    # Trying to log threads while monkey-patched can lead to deadlocks; see
    # https://bugs.launchpad.net/swift/+bug/1895739
    logging.logThreads = 0


class PipeMutex(object):
    """
    Mutex using a pipe. Works across both greenlets and real threads, even
    at the same time.
    """

    def __init__(self):
        self.rfd, self.wfd = os.pipe()

        # You can't create a pipe in non-blocking mode; you must set it
        # later.
        rflags = fcntl.fcntl(self.rfd, fcntl.F_GETFL)
        fcntl.fcntl(self.rfd, fcntl.F_SETFL, rflags | os.O_NONBLOCK)
        os.write(self.wfd, b'-')  # start unlocked

        self.owner = None
        self.recursion_depth = 0

        # Usually, it's an error to have multiple greenthreads all waiting
        # to read the same file descriptor. It's often a sign of inadequate
        # concurrency control; for example, if you have two greenthreads
        # trying to use the same memcache connection, they'll end up writing
        # interleaved garbage to the socket or stealing part of each others'
        # responses.
        #
        # In this case, we have multiple greenthreads waiting on the same
        # file descriptor by design. This lets greenthreads in real thread A
        # wait with greenthreads in real thread B for the same mutex.
        # Therefore, we must turn off eventlet's multiple-reader detection.
        #
        # It would be better to turn off multiple-reader detection for only
        # our calls to trampoline(), but eventlet does not support that.
        eventlet.debug.hub_prevent_multiple_readers(False)

    def acquire(self, blocking=True):
        """
        Acquire the mutex.

        If called with blocking=False, returns True if the mutex was
        acquired and False if it wasn't. Otherwise, blocks until the mutex
        is acquired and returns True.

        This lock is recursive; the same greenthread may acquire it as many
        times as it wants to, though it must then release it that many times
        too.
        """
        current_greenthread_id = id(eventlet.greenthread.getcurrent())
        if self.owner == current_greenthread_id:
            self.recursion_depth += 1
            return True

        while True:
            try:
                # If there is a byte available, this will read it and remove
                # it from the pipe. If not, this will raise OSError with
                # errno=EAGAIN.
                os.read(self.rfd, 1)
                self.owner = current_greenthread_id
                return True
            except OSError as err:
                if err.errno != errno.EAGAIN:
                    raise

                if not blocking:
                    return False

                # Tell eventlet to suspend the current greenthread until
                # self.rfd becomes readable. This will happen when someone
                # else writes to self.wfd.
                eventlet.hubs.trampoline(self.rfd, read=True)

    def release(self):
        """
        Release the mutex.
        """
        current_greenthread_id = id(eventlet.greenthread.getcurrent())
        if self.owner != current_greenthread_id:
            raise RuntimeError("cannot release un-acquired lock")

        if self.recursion_depth > 0:
            self.recursion_depth -= 1
            return

        self.owner = None
        os.write(self.wfd, b'X')

    def close(self):
        """
        Close the mutex. This releases its file descriptors.

        You can't use a mutex after it's been closed.
        """
        if self.wfd is not None:
            os.close(self.rfd)
            self.rfd = None
            os.close(self.wfd)
            self.wfd = None
        self.owner = None
        self.recursion_depth = 0

    def __del__(self):
        # We need this so we don't leak file descriptors. Otherwise, if you
        # call get_swift_logger() and don't explicitly dispose of it by calling
        # logger.logger.handlers[0].lock.close() [1], the pipe file
        # descriptors are leaked.
        #
        # This only really comes up in tests. Swift processes tend to call
        # get_swift_logger() once and then hang on to it until they exit,
        # but the test suite calls get_swift_logger() a lot.
        #
        # [1] and that's a completely ridiculous thing to expect callers to
        # do, so nobody does it and that's okay.
        self.close()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, *args):
        self.release()


class NoopMutex(object):
    """
    "Mutex" that doesn't lock anything.

    We only allow our syslog logging to be configured via UDS or UDP, neither
    of which have the message-interleaving trouble you'd expect from TCP or
    file handlers.
    """

    def __init__(self):
        # Usually, it's an error to have multiple greenthreads all waiting
        # to write to the same file descriptor. It's often a sign of inadequate
        # concurrency control; for example, if you have two greenthreads
        # trying to use the same memcache connection, they'll end up writing
        # interleaved garbage to the socket or stealing part of each others'
        # responses.
        #
        # In this case, we have multiple greenthreads waiting on the same
        # (logging) file descriptor by design. So, similar to the PipeMutex,
        # we must turn off eventlet's multiple-waiter detection.
        #
        # It would be better to turn off multiple-reader detection for only
        # the logging socket fd, but eventlet does not support that.
        eventlet.debug.hub_prevent_multiple_readers(False)

    def acquire(self, blocking=True):
        pass

    def release(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class ThreadSafeSysLogHandler(SysLogHandler):
    def createLock(self):
        if config_true_value(os.environ.get(
                'SWIFT_NOOP_LOGGING_MUTEX') or 'true'):
            self.lock = NoopMutex()
        else:
            self.lock = PipeMutex()


# double inheritance to support property with setter
class SwiftLogAdapter(logging.LoggerAdapter, object):
    """
    A LogAdapter that modifies the adapted ``Logger`` instance
    in the following ways:

    * Performs some reformatting on calls to :meth:`exception`.
    * Provides threadlocal txn_id and client_ip attributes.
    * Adds the txn_id, client_ip and server attributes to the ``extras`` dict
      when a message is processed.
    * Adds the given prefix to the start of each log message.
    * Provides a notice method for logging at NOTICE level.
    """

    _cls_thread_local = threading.local()

    def __init__(self, logger, server, prefix=''):
        logging.LoggerAdapter.__init__(self, logger, {})
        self.prefix = prefix
        self.server = server
        self.warn = self.warning

    @property
    def txn_id(self):
        if hasattr(self._cls_thread_local, 'txn_id'):
            return self._cls_thread_local.txn_id

    @txn_id.setter
    def txn_id(self, value):
        self._cls_thread_local.txn_id = value

    @property
    def client_ip(self):
        if hasattr(self._cls_thread_local, 'client_ip'):
            return self._cls_thread_local.client_ip

    @client_ip.setter
    def client_ip(self, value):
        self._cls_thread_local.client_ip = value

    @property
    def thread_locals(self):
        return (self.txn_id, self.client_ip)

    @thread_locals.setter
    def thread_locals(self, value):
        self.txn_id, self.client_ip = value

    def process(self, msg, kwargs):
        """
        Add extra info to message
        """
        kwargs['extra'] = {'server': self.server, 'txn_id': self.txn_id,
                           'client_ip': self.client_ip}
        msg = '%s%s' % (self.prefix, msg)
        return msg, kwargs

    def notice(self, msg, *args, **kwargs):
        """
        Convenience function for syslog priority LOG_NOTICE. The python
        logging lvl is set to 25, just above info.  SysLogHandler is
        monkey patched to map this log lvl to the LOG_NOTICE syslog
        priority.
        """
        self.log(NOTICE, msg, *args, **kwargs)

    def _exception(self, msg, *args, **kwargs):
        # We up-call to exception() where stdlib uses error() so we can get
        # some of the traceback suppression from LogAdapter, below
        logging.LoggerAdapter.exception(self, msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        _junk, exc, _junk = sys.exc_info()
        call = self.error
        emsg = ''
        if isinstance(exc, (http.client.BadStatusLine,
                            green_http_client.BadStatusLine)):
            # Use error(); not really exceptional
            emsg = repr(exc)
            # Note that on py3, we've seen a RemoteDisconnected error getting
            # raised, which inherits from *both* BadStatusLine and OSError;
            # we want it getting caught here
        elif isinstance(exc, (OSError, socket.error)):
            if exc.errno in (errno.EIO, errno.ENOSPC):
                emsg = str(exc)
            elif exc.errno == errno.ECONNREFUSED:
                emsg = 'Connection refused'
            elif exc.errno == errno.ECONNRESET:
                emsg = 'Connection reset'
            elif exc.errno == errno.EHOSTUNREACH:
                emsg = 'Host unreachable'
            elif exc.errno == errno.ENETUNREACH:
                emsg = 'Network unreachable'
            elif exc.errno == errno.ETIMEDOUT:
                emsg = 'Connection timeout'
            elif exc.errno == errno.EPIPE:
                emsg = 'Broken pipe'
            else:
                call = self._exception
        elif isinstance(exc, eventlet.Timeout):
            emsg = exc.__class__.__name__
            detail = '%ss' % exc.seconds
            if hasattr(exc, 'created_at'):
                detail += ' after %0.2fs' % (time.time() - exc.created_at)
            emsg += ' (%s)' % detail
            if isinstance(exc, swift.common.exceptions.MessageTimeout):
                if exc.msg:
                    emsg += ' %s' % exc.msg
        else:
            call = self._exception
        call('%s: %s' % (msg, emsg), *args, **kwargs)


class SwiftLogFormatter(logging.Formatter):
    """
    Custom logging.Formatter will append txn_id to a log message if the
    record has one and the message does not. Optionally it can shorten
    overly long log lines.
    """

    def __init__(self, fmt=None, datefmt=None, max_line_length=0):
        logging.Formatter.__init__(self, fmt=fmt, datefmt=datefmt)
        self.max_line_length = max_line_length

    def format(self, record):
        if not hasattr(record, 'server'):
            # Catch log messages that were not initiated by swift
            # (for example, the keystone auth middleware)
            record.server = record.name

        # Included from Python's logging.Formatter and then altered slightly to
        # replace \n with #012
        record.message = record.getMessage()
        if self._fmt.find('%(asctime)') >= 0:
            record.asctime = self.formatTime(record, self.datefmt)
        msg = (self._fmt % record.__dict__).replace('\n', '#012')
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(
                    record.exc_info).replace('\n', '#012')
        if record.exc_text:
            if not msg.endswith('#012'):
                msg = msg + '#012'
            msg = msg + record.exc_text

        if (hasattr(record, 'txn_id') and record.txn_id and
                record.txn_id not in msg):
            msg = "%s (txn: %s)" % (msg, record.txn_id)
        if (hasattr(record, 'client_ip') and record.client_ip and
                record.levelno != logging.INFO and
                record.client_ip not in msg):
            msg = "%s (client_ip: %s)" % (msg, record.client_ip)
        if self.max_line_length > 0 and len(msg) > self.max_line_length:
            if self.max_line_length < 7:
                msg = msg[:self.max_line_length]
            else:
                approxhalf = (self.max_line_length - 5) // 2
                msg = msg[:approxhalf] + " ... " + msg[-approxhalf:]
        return msg


class LoggerFileObject(object):

    # Note: this is greenthread-local storage
    _cls_thread_local = threading.local()

    def __init__(self, logger, log_type='STDOUT'):
        self.logger = logger
        self.log_type = log_type

    def write(self, value):
        # We can get into a nasty situation when logs are going to syslog
        # and syslog dies.
        #
        # It's something like this:
        #
        # (A) someone logs something
        #
        # (B) there's an exception in sending to /dev/log since syslog is
        #     not working
        #
        # (C) logging takes that exception and writes it to stderr (see
        #     logging.Handler.handleError)
        #
        # (D) stderr was replaced with a LoggerFileObject at process start,
        #     so the LoggerFileObject takes the provided string and tells
        #     its logger to log it (to syslog, naturally).
        #
        # Then, steps B through D repeat until we run out of stack.
        if getattr(self._cls_thread_local, 'already_called_write', False):
            return

        self._cls_thread_local.already_called_write = True
        try:
            value = value.strip()
            if value:
                if 'Connection reset by peer' in value:
                    self.logger.error(
                        '%s: Connection reset by peer', self.log_type)
                else:
                    self.logger.error('%(type)s: %(value)s',
                                      {'type': self.log_type, 'value': value})
        finally:
            self._cls_thread_local.already_called_write = False

    def writelines(self, values):
        if getattr(self._cls_thread_local, 'already_called_writelines', False):
            return

        self._cls_thread_local.already_called_writelines = True
        try:
            self.logger.error('%(type)s: %(value)s',
                              {'type': self.log_type,
                               'value': '#012'.join(values)})
        finally:
            self._cls_thread_local.already_called_writelines = False

    def close(self):
        pass

    def flush(self):
        pass

    def __iter__(self):
        return self

    def __next__(self):
        raise IOError(errno.EBADF, 'Bad file descriptor')

    def read(self, size=-1):
        raise IOError(errno.EBADF, 'Bad file descriptor')

    def readline(self, size=-1):
        raise IOError(errno.EBADF, 'Bad file descriptor')

    def tell(self):
        return 0

    def xreadlines(self):
        return self


class LogLevelFilter(object):
    """
    Drop messages for the logger based on level.

    This is useful when dependencies log too much information.

    :param level: All messages at or below this level are dropped
                  (DEBUG < INFO < WARN < ERROR < CRITICAL|FATAL)
                  Default: DEBUG
    """

    def __init__(self, level=logging.DEBUG):
        self.level = level

    def filter(self, record):
        if record.levelno <= self.level:
            return 0
        return 1


def get_swift_logger(conf, name=None, log_to_console=False, log_route=None,
                     fmt="%(server)s: %(message)s"):
    """
    Get the current system logger using config settings.

    **Log config and defaults**::

        log_facility = LOG_LOCAL0
        log_level = INFO
        log_name = swift
        log_max_line_length = 0
        log_udp_host = (disabled)
        log_udp_port = logging.handlers.SYSLOG_UDP_PORT
        log_address = /dev/log

    :param conf: Configuration dict to read settings from
    :param name: This value is used to populate the ``server`` field in
                 the log format, as the default value for ``log_route``;
                 defaults to the ``log_name`` value in ``conf``, if it exists,
                 or to 'swift'.
    :param log_to_console: Add handler which writes to console on stderr
    :param log_route: Route for the logging, not emitted to the log, just used
                      to separate logging configurations; defaults to the value
                      of ``name`` or whatever ``name`` defaults to. This value
                      is used as the name attribute of the
                      ``logging.LogAdapter`` that is returned.
    :param fmt: Override log format
    :return: an instance of ``SwiftLogAdapter``
    """
    # note: log_name is typically specified in conf (i.e. defined by
    # operators), whereas log_route is typically hard-coded in callers of
    # get_swift_logger (i.e. defined by developers)
    if not conf:
        conf = {}
    if name is None:
        name = conf.get('log_name', 'swift')
    if not log_route:
        log_route = name
    logger = logging.getLogger(log_route)
    logger.propagate = False
    # all new handlers will get the same formatter
    formatter = SwiftLogFormatter(
        fmt=fmt, max_line_length=int(conf.get('log_max_line_length', 0)))

    # get_swift_logger will only ever add one SysLog Handler to a logger
    if not hasattr(get_swift_logger, 'handler4logger'):
        get_swift_logger.handler4logger = {}
    if logger in get_swift_logger.handler4logger:
        logger.removeHandler(get_swift_logger.handler4logger[logger])

    # facility for this logger will be set by last call wins
    facility = getattr(SysLogHandler, conf.get('log_facility', 'LOG_LOCAL0'),
                       SysLogHandler.LOG_LOCAL0)
    udp_host = conf.get('log_udp_host')
    if udp_host:
        udp_port = int(conf.get('log_udp_port',
                                logging.handlers.SYSLOG_UDP_PORT))
        handler = ThreadSafeSysLogHandler(address=(udp_host, udp_port),
                                          facility=facility)
    else:
        log_address = conf.get('log_address', '/dev/log')
        handler = None
        try:
            mode = os.stat(log_address).st_mode
            if stat.S_ISSOCK(mode):
                handler = ThreadSafeSysLogHandler(address=log_address,
                                                  facility=facility)
        except (OSError, socket.error) as e:
            # If either /dev/log isn't a UNIX socket or it does not exist at
            # all then py2 would raise an error
            if e.errno not in [errno.ENOTSOCK, errno.ENOENT]:
                raise
        if handler is None:
            # fallback to default UDP
            handler = ThreadSafeSysLogHandler(facility=facility)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    get_swift_logger.handler4logger[logger] = handler

    # setup console logging
    if log_to_console or hasattr(get_swift_logger, 'console_handler4logger'):
        # remove pre-existing console handler for this logger
        if not hasattr(get_swift_logger, 'console_handler4logger'):
            get_swift_logger.console_handler4logger = {}
        if logger in get_swift_logger.console_handler4logger:
            logger.removeHandler(
                get_swift_logger.console_handler4logger[logger])

        console_handler = logging.StreamHandler(sys.__stderr__)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
        get_swift_logger.console_handler4logger[logger] = console_handler

    # set the level for the logger
    logger.setLevel(
        getattr(logging, conf.get('log_level', 'INFO').upper(), logging.INFO))

    adapted_logger = SwiftLogAdapter(logger, name)
    other_handlers = conf.get('log_custom_handlers', None)
    if other_handlers:
        log_custom_handlers = [s.strip() for s in other_handlers.split(',')
                               if s.strip()]
        for hook in log_custom_handlers:
            try:
                mod, fnc = hook.rsplit('.', 1)
                logger_hook = getattr(__import__(mod, fromlist=[fnc]), fnc)
                logger_hook(conf, name, log_to_console, log_route, fmt,
                            logger, adapted_logger)
            except (AttributeError, ImportError):
                print('Error calling custom handler [%s]' % hook,
                      file=sys.stderr)
            except ValueError:
                print('Invalid custom handler format [%s]' % hook,
                      file=sys.stderr)

    return adapted_logger


def get_prefixed_swift_logger(swift_logger, prefix):
    """
    Return a clone of the given ``swift_logger`` with a new prefix string
    that replaces the prefix string of the given ``swift_logger``.

    :param swift_logger: an instance of ``SwiftLogAdapter``.
    :param prefix: a string prefix.
    :returns: a new instance of ``SwiftLogAdapter``.
    """
    return SwiftLogAdapter(
        swift_logger.logger, swift_logger.server, prefix=prefix)


class NullLogger(object):
    """A no-op logger for eventlet wsgi."""

    def write(self, *args):
        # "Logs" the args to nowhere
        pass

    def exception(self, *args):
        pass

    def critical(self, *args):
        pass

    def error(self, *args):
        pass

    def warning(self, *args):
        pass

    def info(self, *args):
        pass

    def debug(self, *args):
        pass

    def log(self, *args):
        pass


def capture_stdio(logger, **kwargs):
    """
    Log unhandled exceptions, close stdio, capture stdout and stderr.

    param logger: Logger object to use
    """
    # log uncaught exceptions
    sys.excepthook = lambda * exc_info: \
        logger.critical('UNCAUGHT EXCEPTION', exc_info=exc_info)

    # collect stdio file desc not in use for logging
    stdio_files = [sys.stdin, sys.stdout, sys.stderr]
    console_fds = [h.stream.fileno() for _junk, h in getattr(
        get_swift_logger, 'console_handler4logger', {}).items()]
    stdio_files = [f for f in stdio_files if f.fileno() not in console_fds]

    with open(os.devnull, 'r+b') as nullfile:
        # close stdio (excludes fds open for logging)
        for f in stdio_files:
            # some platforms throw an error when attempting an stdin flush
            try:
                f.flush()
            except IOError:
                pass

            try:
                os.dup2(nullfile.fileno(), f.fileno())
            except OSError:
                pass

    # redirect stdio
    if kwargs.pop('capture_stdout', True):
        sys.stdout = LoggerFileObject(logger)
    if kwargs.pop('capture_stderr', True):
        sys.stderr = LoggerFileObject(logger, 'STDERR')


class StrAnonymizer(str):
    """
    Class that permits to get a string anonymized or simply quoted.
    """

    def __new__(cls, data, method, salt):
        method = method.lower()
        if method not in hashlib.algorithms_guaranteed:
            raise ValueError('Unsupported hashing method: %r' % method)
        s = str.__new__(cls, data or '')
        s.method = method
        s.salt = salt
        return s

    @property
    def anonymized(self):
        if not self:
            return self
        else:
            if self.method == 'md5':
                h = md5(usedforsecurity=False)
            else:
                h = getattr(hashlib, self.method)()
            if self.salt:
                h.update(self.salt.encode('latin1'))
            h.update(self.encode('latin1'))
            return '{%s%s}%s' % ('S' if self.salt else '', self.method.upper(),
                                 h.hexdigest())


class StrFormatTime(object):
    """
    Class that permits to get formats or parts of a time.
    """

    def __init__(self, ts):
        self.time = ts
        self.time_struct = time.gmtime(ts)

    def __str__(self):
        return "%.9f" % self.time

    def __getattr__(self, attr):
        if attr not in ['a', 'A', 'b', 'B', 'c', 'd', 'H',
                        'I', 'j', 'm', 'M', 'p', 'S', 'U',
                        'w', 'W', 'x', 'X', 'y', 'Y', 'Z']:
            raise ValueError(("The attribute %s is not a correct directive "
                              "for time.strftime formater.") % attr)
        return datetime.datetime(*self.time_struct[:-2],
                                 tzinfo=UTC).strftime('%' + attr)

    @property
    def asctime(self):
        return time.asctime(self.time_struct)

    @property
    def datetime(self):
        return time.strftime('%d/%b/%Y/%H/%M/%S', self.time_struct)

    @property
    def iso8601(self):
        return time.strftime('%Y-%m-%dT%H:%M:%S', self.time_struct)

    @property
    def ms(self):
        return self.__str__().split('.')[1][:3]

    @property
    def us(self):
        return self.__str__().split('.')[1][:6]

    @property
    def ns(self):
        return self.__str__().split('.')[1]

    @property
    def s(self):
        return self.__str__().split('.')[0]


def get_log_line(req, res, trans_time, additional_info, fmt,
                 anonymization_method, anonymization_salt):
    """
    Make a line for logging that matches the documented log line format
    for backend servers.

    :param req: the request.
    :param res: the response.
    :param trans_time: the time the request took to complete, a float.
    :param additional_info: a string to log at the end of the line

    :returns: a properly formatted line for logging.
    """

    policy_index = get_policy_index(req.headers, res.headers)
    if req.path.startswith('/'):
        disk, partition, account, container, obj = split_path(req.path, 0, 5,
                                                              True)
    else:
        disk, partition, account, container, obj = (None, ) * 5
    replacements = {
        'remote_addr': StrAnonymizer(req.remote_addr, anonymization_method,
                                     anonymization_salt),
        'time': StrFormatTime(time.time()),
        'method': req.method,
        'path': StrAnonymizer(req.path, anonymization_method,
                              anonymization_salt),
        'disk': disk,
        'partition': partition,
        'account': StrAnonymizer(account, anonymization_method,
                                 anonymization_salt),
        'container': StrAnonymizer(container, anonymization_method,
                                   anonymization_salt),
        'object': StrAnonymizer(obj, anonymization_method,
                                anonymization_salt),
        'status': res.status.split()[0],
        'content_length': res.content_length,
        'referer': StrAnonymizer(req.referer, anonymization_method,
                                 anonymization_salt),
        'txn_id': req.headers.get('x-trans-id'),
        'user_agent': StrAnonymizer(req.user_agent, anonymization_method,
                                    anonymization_salt),
        'trans_time': trans_time,
        'additional_info': additional_info,
        'pid': os.getpid(),
        'policy_index': policy_index,
    }
    return LogStringFormatter(default='-').format(fmt, **replacements)


def get_policy_index(req_headers, res_headers):
    """
    Returns the appropriate index of the storage policy for the request from
    a proxy server

    :param req_headers: dict of the request headers.
    :param res_headers: dict of the response headers.

    :returns: string index of storage policy, or None
    """
    header = 'X-Backend-Storage-Policy-Index'
    policy_index = res_headers.get(header, req_headers.get(header))
    if isinstance(policy_index, bytes):
        policy_index = policy_index.decode('ascii')
    return str(policy_index) if policy_index is not None else None


class LogStringFormatter(string.Formatter):
    def __init__(self, default='', quote=False):
        super(LogStringFormatter, self).__init__()
        self.default = default
        self.quote = quote

    def format_field(self, value, spec):
        if not value:
            return self.default
        else:
            log = super(LogStringFormatter, self).format_field(value, spec)
            if self.quote:
                return quote(log, ':/{}')
            else:
                return log



# Copyright (c) 2010-2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json

from swift.common.constraints import MAX_OBJECT_NAME_LENGTH
from swift.common.http import HTTP_NO_CONTENT
from swift.common.swob import str_to_wsgi
from swift.common.utils import public, StreamingPile
from swift.common.registry import get_swift_info

from swift.common.middleware.s3api.controllers.base import Controller, \
    bucket_operation
from swift.common.middleware.s3api.etree import Element, SubElement, \
    fromstring, tostring, XMLSyntaxError, DocumentInvalid
from swift.common.middleware.s3api.s3response import HTTPOk, \
    S3NotImplemented, NoSuchKey, ErrorResponse, MalformedXML, \
    UserKeyMustBeSpecified, AccessDenied, MissingRequestBodyError


class MultiObjectDeleteController(Controller):
    """
    Handles Delete Multiple Objects, which is logged as a MULTI_OBJECT_DELETE
    operation in the S3 server log.
    """
    def _gen_error_body(self, error, elem, delete_list):
        for key, version in delete_list:
            error_elem = SubElement(elem, 'Error')
            SubElement(error_elem, 'Key').text = key
            if version is not None:
                SubElement(error_elem, 'VersionId').text = version
            SubElement(error_elem, 'Code').text = error.__class__.__name__
            SubElement(error_elem, 'Message').text = error._msg

        return tostring(elem)

    @public
    @bucket_operation
    def POST(self, req):
        """
        Handles Delete Multiple Objects.
        """
        def object_key_iter(elem):
            for obj in elem.iterchildren('Object'):
                key = obj.find('./Key').text
                if not key:
                    raise UserKeyMustBeSpecified()
                version = obj.find('./VersionId')
                if version is not None:
                    version = version.text

                yield key, version

        max_body_size = min(
            # FWIW, AWS limits multideletes to 1000 keys, and swift limits
            # object names to 1024 bytes (by default). Add a factor of two to
            # allow some slop.
            2 * self.conf.max_multi_delete_objects * MAX_OBJECT_NAME_LENGTH,
            # But, don't let operators shoot themselves in the foot
            10 * 1024 * 1024)

        try:
            xml = req.xml(max_body_size)
            if not xml:
                raise MissingRequestBodyError()

            if 'x-amz-content-sha256' not in req.headers:
                # SHA256 got checked when we read the body, so there's at
                # least *some* verification. Recent versions of boto3 stopped
                # sending Content-MD5, so it can't *always* be required.
                # See https://bugs.launchpad.net/swift/+bug/2098529
                req.check_md5(xml)
            elem = fromstring(xml, 'Delete', self.logger)

            quiet = elem.find('./Quiet')
            self.quiet = quiet is not None and quiet.text.lower() == 'true'

            delete_list = list(object_key_iter(elem))
            if len(delete_list) > self.conf.max_multi_delete_objects:
                raise MalformedXML()
        except (XMLSyntaxError, DocumentInvalid):
            raise MalformedXML()
        except ErrorResponse:
            raise
        except Exception as e:
            self.logger.error(e)
            raise

        elem = Element('DeleteResult')

        # check bucket existence
        try:
            req.get_response(self.app, 'HEAD')
        except AccessDenied as error:
            body = self._gen_error_body(error, elem, delete_list)
            return HTTPOk(body=body)

        if 'object_versioning' not in get_swift_info() and any(
                version not in ('null', None)
                for _key, version in delete_list):
            raise S3NotImplemented()

        def do_delete(base_req, key, version):
            req = copy.copy(base_req)
            req.environ = copy.copy(base_req.environ)
            req.object_name = str_to_wsgi(key)
            if version:
                req.params = {'version-id': version, 'symlink': 'get'}

            try:
                try:
                    query = req.gen_multipart_manifest_delete_query(
                        self.app, version=version)
                except NoSuchKey:
                    query = {}
                if version:
                    query['version-id'] = version
                    query['symlink'] = 'get'

                resp = req.get_response(self.app, method='DELETE', query=query,
                                        headers={'Accept': 'application/json'})
                # If async segment cleanup is available, we expect to get
                # back a 204; otherwise, the delete is synchronous and we
                # have to read the response to actually do the SLO delete
                if query.get('multipart-manifest') and \
                        resp.status_int != HTTP_NO_CONTENT:
                    try:
                        delete_result = json.loads(resp.body)
                        if delete_result['Errors']:
                            # NB: bulk includes 404s in "Number Not Found",
                            # not "Errors"
                            msg_parts = [delete_result['Response Status']]
                            msg_parts.extend(
                                '%s: %s' % (obj, status)
                                for obj, status in delete_result['Errors'])
                            return key, {'code': 'SLODeleteError',
                                         'message': '\n'.join(msg_parts)}
                        # else, all good
                    except (ValueError, TypeError, KeyError):
                        # Logs get all the gory details
                        self.logger.exception(
                            'Could not parse SLO delete response (%s): %s',
                            resp.status, resp.body)
                        # Client gets something more generic
                        return key, {'code': 'SLODeleteError',
                                     'message': 'Unexpected swift response'}
            except NoSuchKey:
                pass
            except ErrorResponse as e:
                return key, {'code': e.__class__.__name__, 'message': e._msg}
            except Exception:
                self.logger.exception(
                    'Unexpected Error handling DELETE of %r %r' % (
                        req.container_name, key))
                return key, {'code': 'Server Error', 'message': 'Server Error'}

            return key, None

        with StreamingPile(self.conf.multi_delete_concurrency) as pile:
            for key, err in pile.asyncstarmap(do_delete, (
                    (req, key, version) for key, version in delete_list)):
                if err:
                    error = SubElement(elem, 'Error')
                    SubElement(error, 'Key').text = key
                    SubElement(error, 'Code').text = err['code']
                    SubElement(error, 'Message').text = err['message']
                elif not self.quiet:
                    deleted = SubElement(elem, 'Deleted')
                    SubElement(deleted, 'Key').text = key

        body = tostring(elem)

        return HTTPOk(body=body)



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Why our own memcache client?
By Michael Barton

python-memcached doesn't use consistent hashing, so adding or
removing a memcache server from the pool invalidates a huge
percentage of cached items.

If you keep a pool of python-memcached client objects, each client
object has its own connection to every memcached server, only one of
which is ever in use.  So you wind up with n * m open sockets and
almost all of them idle. This client effectively has a pool for each
server, so the number of backend connections is hopefully greatly
reduced.

python-memcache uses pickle to store things, and there was already a
huge stink about Swift using pickles in memcache
(http://osvdb.org/show/osvdb/86581).  That seemed sort of unfair,
since nova and keystone and everyone else use pickles for memcache
too, but it's hidden behind a "standard" library. But changing would
be a security regression at this point.

Also, pylibmc wouldn't work for us because it needs to use python
sockets in order to play nice with eventlet.

Lucid comes with memcached: v1.4.2.  Protocol documentation for that
version is at:

http://github.com/memcached/memcached/blob/1.4.2/doc/protocol.txt
"""

import os
import json
import logging
# the name of 'time' module is changed to 'tm', to avoid changing the
# signatures of member functions in this file.
import time as tm
from bisect import bisect

from eventlet.green import socket, ssl
from eventlet.pools import Pool
from eventlet import Timeout
from configparser import ConfigParser, NoSectionError, NoOptionError
from swift.common import utils
from swift.common.exceptions import MemcacheConnectionError, \
    MemcacheIncrNotFoundError, MemcachePoolTimeout
from swift.common.utils import md5, human_readable, config_true_value, \
    memcached_timing_stats

DEFAULT_MEMCACHED_PORT = 11211

CONN_TIMEOUT = 0.3
POOL_TIMEOUT = 1.0  # WAG
IO_TIMEOUT = 2.0
PICKLE_FLAG = 1
JSON_FLAG = 2
NODE_WEIGHT = 50
TRY_COUNT = 3

# if ERROR_LIMIT_COUNT errors occur in ERROR_LIMIT_TIME seconds, the server
# will be considered failed for ERROR_LIMIT_DURATION seconds.
ERROR_LIMIT_COUNT = 10
ERROR_LIMIT_TIME = ERROR_LIMIT_DURATION = 60
DEFAULT_ITEM_SIZE_WARNING_THRESHOLD = -1

# Different sample rates for emitting Memcached timing stats.
TIMING_SAMPLE_RATE_HIGH = 0.1
TIMING_SAMPLE_RATE_MEDIUM = 0.01
TIMING_SAMPLE_RATE_LOW = 0.001

# The max value of a delta expiration time.
EXPTIME_MAXDELTA = 30 * 24 * 60 * 60


def md5hash(key):
    if not isinstance(key, bytes):
        key = key.encode('utf-8', errors='surrogateescape')
    return md5(key, usedforsecurity=False).hexdigest().encode('ascii')


def sanitize_timeout(timeout):
    """
    Sanitize a timeout value to use an absolute expiration time if the delta
    is greater than 30 days (in seconds). Note that the memcached server
    translates negative values to mean a delta of 30 days in seconds (and 1
    additional second), client beware.
    """
    if timeout > EXPTIME_MAXDELTA:
        timeout += tm.time()
    return int(timeout)


def set_msg(key, flags, timeout, value):
    if not isinstance(key, bytes):
        raise TypeError('key must be bytes')
    if not isinstance(value, bytes):
        raise TypeError('value must be bytes')
    return b' '.join([
        b'set',
        key,
        str(flags).encode('ascii'),
        str(timeout).encode('ascii'),
        str(len(value)).encode('ascii'),
    ]) + (b'\r\n' + value + b'\r\n')


class MemcacheConnPool(Pool):
    """
    Connection pool for Memcache Connections

    The *server* parameter can be a hostname, an IPv4 address, or an IPv6
    address with an optional port. See
    :func:`swift.common.utils.parse_socket_string` for details.
    """

    def __init__(self, server, size, connect_timeout, tls_context=None):
        Pool.__init__(self, max_size=size)
        self.host, self.port = utils.parse_socket_string(
            server, DEFAULT_MEMCACHED_PORT)
        self._connect_timeout = connect_timeout
        self._tls_context = tls_context

    def create(self):
        addrs = socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC,
                                   socket.SOCK_STREAM)
        family, socktype, proto, canonname, sockaddr = addrs[0]
        sock = socket.socket(family, socket.SOCK_STREAM)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        try:
            with Timeout(self._connect_timeout):
                sock.connect(sockaddr)
            if self._tls_context:
                sock = self._tls_context.wrap_socket(sock,
                                                     server_hostname=self.host)
        except (Exception, Timeout):
            sock.close()
            raise
        return (sock.makefile('rwb'), sock)

    def get(self):
        fp, sock = super(MemcacheConnPool, self).get()
        try:
            if fp is None:
                # An error happened previously, so we need a new connection
                fp, sock = self.create()
            return fp, sock
        except MemcachePoolTimeout:
            # This is the only place that knows an item was successfully taken
            # from the pool, so it has to be responsible for repopulating it.
            # Any other errors should get handled in _get_conns(); see the
            # comment about timeouts during create() there.
            self.put((None, None))
            raise


class MemcacheCommand(object):
    """
    Helper class that encapsulates common parameters of a command.

    :param method: the name of the MemcacheRing method that was called.
    :param key: the memcached key.
    """
    __slots__ = ('method', 'key', 'command', 'hash_key')

    def __init__(self, method, key):
        self.method = method
        self.key = key
        self.command = method.encode()
        self.hash_key = md5hash(key)

    @property
    def key_prefix(self):
        # get the prefix of a user provided memcache key by removing the
        # content after the last '/', all current usages within swift are using
        # prefix, such as "shard-updating-v2", "nvratelimit" and etc.
        return self.key.rsplit('/', 1)[0]


class MemcacheRing(object):
    """
    Simple, consistent-hashed memcache client.
    """

    def __init__(
            self, servers, connect_timeout=CONN_TIMEOUT,
            io_timeout=IO_TIMEOUT, pool_timeout=POOL_TIMEOUT,
            tries=TRY_COUNT,
            max_conns=2, tls_context=None, logger=None,
            error_limit_count=ERROR_LIMIT_COUNT,
            error_limit_time=ERROR_LIMIT_TIME,
            error_limit_duration=ERROR_LIMIT_DURATION,
            item_size_warning_threshold=DEFAULT_ITEM_SIZE_WARNING_THRESHOLD):
        self._ring = {}
        self._errors = dict(((serv, []) for serv in servers))
        self._error_limited = dict(((serv, 0) for serv in servers))
        self._error_limit_count = error_limit_count
        self._error_limit_time = error_limit_time
        self._error_limit_duration = error_limit_duration
        for server in sorted(servers):
            for i in range(NODE_WEIGHT):
                self._ring[md5hash('%s-%s' % (server, i))] = server
        self._tries = tries if tries <= len(servers) else len(servers)
        self._sorted = sorted(self._ring)
        self._client_cache = dict((
            (server, MemcacheConnPool(server, max_conns, connect_timeout,
                                      tls_context=tls_context))
            for server in servers))
        self._connect_timeout = connect_timeout
        self._io_timeout = io_timeout
        self._pool_timeout = pool_timeout
        if logger is None:
            self.logger = logging.getLogger()
        else:
            self.logger = logger
        self.item_size_warning_threshold = item_size_warning_threshold

    @property
    def memcache_servers(self):
        return list(self._client_cache.keys())

    def _log_error(self, server, cmd, action, msg):
        self.logger.error(
            "Error %(action)s to memcached: %(server)s"
            ": with key_prefix %(key_prefix)s, method %(method)s: %(msg)s",
            {'action': action, 'server': server, 'key_prefix': cmd.key_prefix,
             'method': cmd.method, 'msg': msg})

    """
    Handles exceptions.

    :param server: a server.
    :param e: an exception.
    :param cmd: an instance of MemcacheCommand.
    :param conn_start_time: the time at which the failed operation started.
    :param action: a verb describing the operation.
    :param sock: an optional socket that needs to be closed by this method.
    :param fp: an optional file pointer that needs to be closed by this method.
    :param got_connection: if ``True``, the server's connection will be reset
        in the cached connection pool.
    """
    def _exception_occurred(self, server, e, cmd, conn_start_time,
                            action='talking', sock=None,
                            fp=None, got_connection=True):
        if isinstance(e, Timeout):
            self.logger.error(
                "Timeout %(action)s to memcached: %(server)s"
                ": with key_prefix %(key_prefix)s, method %(method)s, "
                "config_timeout %(config_timeout)s, time_spent %(time_spent)s",
                {'action': action, 'server': server,
                 'key_prefix': cmd.key_prefix, 'method': cmd.method,
                 'config_timeout': e.seconds,
                 'time_spent': tm.time() - conn_start_time})
            self.logger.timing_since(
                'memcached.' + cmd.method + '.timeout.timing',
                conn_start_time)
        elif isinstance(e, (socket.error, MemcacheConnectionError)):
            self.logger.error(
                "Error %(action)s to memcached: %(server)s: "
                "with key_prefix %(key_prefix)s, method %(method)s, "
                "time_spent %(time_spent)s, %(err)s",
                {'action': action, 'server': server,
                 'key_prefix': cmd.key_prefix, 'method': cmd.method,
                 'time_spent': tm.time() - conn_start_time, 'err': e})
            self.logger.timing_since(
                'memcached.' + cmd.method + '.conn_err.timing',
                conn_start_time)
        else:
            self.logger.exception(
                "Error %(action)s to memcached: %(server)s"
                ": with key_prefix %(key_prefix)s, method %(method)s, "
                "time_spent %(time_spent)s",
                {'action': action, 'server': server,
                 'key_prefix': cmd.key_prefix, 'method': cmd.method,
                 'time_spent': tm.time() - conn_start_time})
            self.logger.timing_since(
                'memcached.' + cmd.method + '.errors.timing', conn_start_time)

        try:
            if fp:
                fp.close()
                del fp
        except Exception:
            pass
        try:
            if sock:
                sock.close()
                del sock
        except Exception:
            pass
        if got_connection:
            # We need to return something to the pool
            # A new connection will be created the next time it is retrieved
            self._return_conn(server, None, None)

        if isinstance(e, MemcacheIncrNotFoundError):
            # these errors can be caused by other greenthreads not yielding to
            # the incr greenthread often enough, rather than a server problem,
            # so don't error limit the server
            return

        if self._error_limit_time <= 0 or self._error_limit_duration <= 0:
            return

        now = tm.time()
        self._errors[server].append(now)
        if len(self._errors[server]) > self._error_limit_count:
            self._errors[server] = [err for err in self._errors[server]
                                    if err > now - self._error_limit_time]
            if len(self._errors[server]) > self._error_limit_count:
                self._error_limited[server] = now + self._error_limit_duration
                self.logger.error('Error limiting server %s', server)

    def _get_conns(self, cmd):
        """
        Retrieves a server conn from the pool, or connects a new one.
        Chooses the server based on a consistent hash of "key".

        :param cmd: an instance of MemcacheCommand.
        :return: generator to serve memcached connection
        """
        pos = bisect(self._sorted, cmd.hash_key)
        served = []
        any_yielded = False
        while len(served) < self._tries:
            pos = (pos + 1) % len(self._sorted)
            server = self._ring[self._sorted[pos]]
            if server in served:
                continue
            served.append(server)
            pool_start_time = tm.time()
            if self._error_limited[server] > pool_start_time:
                continue
            sock = None
            try:
                with MemcachePoolTimeout(self._pool_timeout):
                    fp, sock = self._client_cache[server].get()
                any_yielded = True
                yield server, fp, sock
            except MemcachePoolTimeout as e:
                self._exception_occurred(server, e, cmd, pool_start_time,
                                         action='getting a connection',
                                         got_connection=False)
            except (Exception, Timeout) as e:
                # Typically a Timeout exception caught here is the one raised
                # by the create() method of this server's MemcacheConnPool
                # object.
                self._exception_occurred(server, e, cmd, pool_start_time,
                                         action='connecting', sock=sock)
        if not any_yielded:
            self._log_error('ALL', cmd, 'connecting',
                            'No more memcached servers to try')

    def _return_conn(self, server, fp, sock):
        """Returns a server connection to the pool."""
        self._client_cache[server].put((fp, sock))

    # Sample rates of different memcached operations are based on generic
    # swift usage patterns.
    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_HIGH)
    def set(self, key, value, serialize=True, time=0,
            min_compress_len=0, raise_on_error=False):
        """
        Set a key/value pair in memcache

        :param key: key
        :param value: value
        :param serialize: if True, value is serialized with JSON before sending
                          to memcache
        :param time: the time to live
        :param min_compress_len: minimum compress length, this parameter was
                                 added to keep the signature compatible with
                                 python-memcached interface. This
                                 implementation ignores it.
        :param raise_on_error: if True, propagate Timeouts and other errors.
                               By default, errors are ignored.
        """
        cmd = MemcacheCommand('set', key)
        timeout = sanitize_timeout(time)
        flags = 0
        if serialize:
            if isinstance(value, bytes):
                value = value.decode('utf8')
            value = json.dumps(value).encode('ascii')
            flags |= JSON_FLAG
        elif not isinstance(value, bytes):
            value = str(value).encode('utf-8')

        if 0 <= self.item_size_warning_threshold <= len(value):
            self.logger.warning(
                "Item size larger than warning threshold: "
                "%d (%s) >= %d (%s)", len(value),
                human_readable(len(value)),
                self.item_size_warning_threshold,
                human_readable(self.item_size_warning_threshold))

        for (server, fp, sock) in self._get_conns(cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    sock.sendall(set_msg(cmd.hash_key, flags, timeout, value))
                    # Wait for the set to complete
                    msg = fp.readline().strip()
                    if msg != b'STORED':
                        msg = msg.decode('ascii')
                        raise MemcacheConnectionError('failed set: %s' % msg)
                    self._return_conn(server, fp, sock)
                    return
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)
        if raise_on_error:
            raise MemcacheConnectionError(
                "No memcached connections succeeded.")

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_MEDIUM)
    def get(self, key, raise_on_error=False):
        """
        Gets the object specified by key.  It will also unserialize the object
        before returning if it is serialized in memcache with JSON.

        :param key: key
        :param raise_on_error: if True, propagate Timeouts and other errors.
                               By default, errors are treated as cache misses.
        :returns: value of the key in memcache
        """
        cmd = MemcacheCommand('get', key)
        value = None
        for (server, fp, sock) in self._get_conns(cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    sock.sendall(b'get ' + cmd.hash_key + b'\r\n')
                    line = fp.readline().strip().split()
                    while True:
                        if not line:
                            raise MemcacheConnectionError('incomplete read')
                        if line[0].upper() == b'END':
                            break
                        if (line[0].upper() == b'VALUE' and
                                line[1] == cmd.hash_key):
                            size = int(line[3])
                            value = fp.read(size)
                            if int(line[2]) & PICKLE_FLAG:
                                value = None
                            if int(line[2]) & JSON_FLAG:
                                value = json.loads(value)
                            fp.readline()
                        line = fp.readline().strip().split()
                    self._return_conn(server, fp, sock)
                    return value
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)
        if raise_on_error:
            raise MemcacheConnectionError(
                "No memcached connections succeeded.")

    def _incr_or_decr(self, fp, sock, cmd, delta):
        sock.sendall(b' '.join([cmd.command, cmd.hash_key, delta]) + b'\r\n')
        line = fp.readline().strip().split()
        if not line:
            raise MemcacheConnectionError('incomplete read')
        if line[0].upper() == b'NOT_FOUND':
            return None
        return int(line[0].strip())

    def _add(self, fp, sock, cmd, add_val, timeout):
        sock.sendall(b' '.join([
            b'add', cmd.hash_key, b'0', str(timeout).encode('ascii'),
            str(len(add_val)).encode('ascii')
        ]) + b'\r\n' + add_val + b'\r\n')
        line = fp.readline().strip().split()
        return None if line[0].upper() == b'NOT_STORED' else int(add_val)

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_LOW)
    def incr(self, key, delta=1, time=0):
        """
        Increments a key which has a numeric value by delta.
        If the key can't be found, it's added as delta or 0 if delta < 0.
        If passed a negative number, will use memcached's decr. Returns
        the int stored in memcached
        Note: The data memcached stores as the result of incr/decr is
        an unsigned int.  decr's that result in a number below 0 are
        stored as 0.

        :param key: key
        :param delta: amount to add to the value of key (or set as the value
                      if the key is not found) will be cast to an int
        :param time: the time to live
        :returns: result of incrementing
        :raises MemcacheConnectionError:
        """
        cmd = MemcacheCommand('incr' if delta >= 0 else 'decr', key)
        delta_val = str(abs(int(delta))).encode('ascii')
        timeout = sanitize_timeout(time)
        for (server, fp, sock) in self._get_conns(cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    new_val = self._incr_or_decr(fp, sock, cmd, delta_val)
                    if new_val is None:
                        add_val = b'0' if cmd.method == 'decr' else delta_val
                        new_val = self._add(fp, sock, cmd, add_val, timeout)
                        if new_val is None:
                            new_val = self._incr_or_decr(
                                fp, sock, cmd, delta_val)
                            if new_val is None:
                                # This can happen if this thread takes more
                                # than the TTL to get from the first failed
                                # incr to the second incr, during which time
                                # the key was concurrently added and expired.
                                raise MemcacheIncrNotFoundError(
                                    'expired ttl=%s' % time)
                    self._return_conn(server, fp, sock)
                    return new_val
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)
        raise MemcacheConnectionError("No memcached connections succeeded.")

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_LOW)
    def decr(self, key, delta=1, time=0):
        """
        Decrements a key which has a numeric value by delta. Calls incr with
        -delta.

        :param key: key
        :param delta: amount to subtract to the value of key (or set the
                      value to 0 if the key is not found) will be cast to
                      an int
        :param time: the time to live
        :returns: result of decrementing
        :raises MemcacheConnectionError:
        """
        return self.incr(key, delta=-delta, time=time)

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_HIGH)
    def delete(self, key, server_key=None):
        """
        Deletes a key/value pair from memcache.

        :param key: key to be deleted
        :param server_key: key to use in determining which server in the ring
                            is used
        """
        cmd = server_cmd = MemcacheCommand('delete', key)
        if server_key:
            server_cmd = MemcacheCommand('delete', server_key)
        for (server, fp, sock) in self._get_conns(server_cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    sock.sendall(b'delete ' + cmd.hash_key + b'\r\n')
                    # Wait for the delete to complete
                    fp.readline()
                    self._return_conn(server, fp, sock)
                    return
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_HIGH)
    def set_multi(self, mapping, server_key, serialize=True, time=0,
                  min_compress_len=0):
        """
        Sets multiple key/value pairs in memcache.

        :param mapping: dictionary of keys and values to be set in memcache
        :param server_key: key to use in determining which server in the ring
                            is used
        :param serialize: if True, value is serialized with JSON before sending
                          to memcache.
        :param time: the time to live
        :min_compress_len: minimum compress length, this parameter was added
                           to keep the signature compatible with
                           python-memcached interface. This implementation
                           ignores it
        """
        cmd = MemcacheCommand('set_multi', server_key)
        timeout = sanitize_timeout(time)
        msg = []
        for key, value in mapping.items():
            key = md5hash(key)
            flags = 0
            if serialize:
                if isinstance(value, bytes):
                    value = value.decode('utf8')
                value = json.dumps(value).encode('ascii')
                flags |= JSON_FLAG
            msg.append(set_msg(key, flags, timeout, value))
        for (server, fp, sock) in self._get_conns(cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    sock.sendall(b''.join(msg))
                    # Wait for the set to complete
                    for line in range(len(mapping)):
                        fp.readline()
                    self._return_conn(server, fp, sock)
                    return
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)

    @memcached_timing_stats(sample_rate=TIMING_SAMPLE_RATE_HIGH)
    def get_multi(self, keys, server_key):
        """
        Gets multiple values from memcache for the given keys.

        :param keys: keys for values to be retrieved from memcache
        :param server_key: key to use in determining which server in the ring
                           is used
        :returns: list of values
        """
        cmd = MemcacheCommand('get_multi', server_key)
        hash_keys = [md5hash(key) for key in keys]
        for (server, fp, sock) in self._get_conns(cmd):
            conn_start_time = tm.time()
            try:
                with Timeout(self._io_timeout):
                    sock.sendall(b'get ' + b' '.join(hash_keys) + b'\r\n')
                    line = fp.readline().strip().split()
                    responses = {}
                    while True:
                        if not line:
                            raise MemcacheConnectionError('incomplete read')
                        if line[0].upper() == b'END':
                            break
                        if line[0].upper() == b'VALUE':
                            size = int(line[3])
                            value = fp.read(size)
                            if int(line[2]) & PICKLE_FLAG:
                                value = None
                            elif int(line[2]) & JSON_FLAG:
                                value = json.loads(value)
                            responses[line[1]] = value
                            fp.readline()
                        line = fp.readline().strip().split()
                    values = []
                    for key in hash_keys:
                        if key in responses:
                            values.append(responses[key])
                        else:
                            values.append(None)
                    self._return_conn(server, fp, sock)
                    return values
            except (Exception, Timeout) as e:
                self._exception_occurred(server, e, cmd, conn_start_time,
                                         sock=sock, fp=fp)


def load_memcache(conf, logger):
    """
    Build a MemcacheRing object from the given config.  It will also use the
    passed in logger.

    :param conf: a dict, the config options
    :param logger: a logger
    """
    memcache_servers = conf.get('memcache_servers')
    try:
        # Originally, while we documented using memcache_max_connections
        # we only accepted max_connections
        max_conns = int(conf.get('memcache_max_connections',
                                 conf.get('max_connections', 0)))
    except ValueError:
        max_conns = 0

    memcache_options = {}
    if (not memcache_servers
            or max_conns <= 0):
        path = os.path.join(conf.get('swift_dir', '/etc/swift'),
                            'memcache.conf')
        memcache_conf = ConfigParser()
        if memcache_conf.read(path):
            # if memcache.conf exists we'll start with those base options
            try:
                memcache_options = dict(memcache_conf.items('memcache'))
            except NoSectionError:
                pass

            if not memcache_servers:
                try:
                    memcache_servers = \
                        memcache_conf.get('memcache', 'memcache_servers')
                except (NoSectionError, NoOptionError):
                    pass
            if max_conns <= 0:
                try:
                    new_max_conns = \
                        memcache_conf.get('memcache',
                                          'memcache_max_connections')
                    max_conns = int(new_max_conns)
                except (NoSectionError, NoOptionError, ValueError):
                    pass

    # while memcache.conf options are the base for the memcache
    # middleware, if you set the same option also in the filter
    # section of the proxy config it is more specific.
    memcache_options.update(conf)
    connect_timeout = float(memcache_options.get(
        'connect_timeout', CONN_TIMEOUT))
    pool_timeout = float(memcache_options.get(
        'pool_timeout', POOL_TIMEOUT))
    tries = int(memcache_options.get('tries', TRY_COUNT))
    io_timeout = float(memcache_options.get('io_timeout', IO_TIMEOUT))
    if config_true_value(memcache_options.get('tls_enabled', 'false')):
        tls_cafile = memcache_options.get('tls_cafile')
        tls_certfile = memcache_options.get('tls_certfile')
        tls_keyfile = memcache_options.get('tls_keyfile')
        tls_context = ssl.create_default_context(
            cafile=tls_cafile)
        if tls_certfile:
            tls_context.load_cert_chain(tls_certfile, tls_keyfile)
    else:
        tls_context = None
    error_suppression_interval = float(memcache_options.get(
        'error_suppression_interval', ERROR_LIMIT_TIME))
    error_suppression_limit = float(memcache_options.get(
        'error_suppression_limit', ERROR_LIMIT_COUNT))
    item_size_warning_threshold = int(memcache_options.get(
        'item_size_warning_threshold', DEFAULT_ITEM_SIZE_WARNING_THRESHOLD))

    if not memcache_servers:
        memcache_servers = '127.0.0.1:11211'
    if max_conns <= 0:
        max_conns = 2

    return MemcacheRing(
        [s.strip() for s in memcache_servers.split(',')
         if s.strip()],
        connect_timeout=connect_timeout,
        pool_timeout=pool_timeout,
        tries=tries,
        io_timeout=io_timeout,
        max_conns=max_conns,
        tls_context=tls_context,
        logger=logger,
        error_limit_count=error_suppression_limit,
        error_limit_time=error_suppression_interval,
        error_limit_duration=error_suppression_interval,
        item_size_warning_threshold=item_size_warning_threshold)



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock
import unittest

from swift.common.middleware import read_only
from swift.common.swob import Request
from test.debug_logger import debug_logger


class FakeApp(object):
    def __call__(self, env, start_response):
        start_response('200 OK', [])
        return [b'Some Content']


def start_response(*args):
    pass


read_methods = 'GET HEAD'.split()
write_methods = 'COPY DELETE POST PUT'.split()
ro_resp = [b'Writes are disabled for this account.']


class TestReadOnly(unittest.TestCase):
    def test_global_read_only_off(self):
        conf = {
            'read_only': 'false',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={}):
            for method in read_methods + write_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_on(self):
        conf = {
            'read_only': 'true',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={}):
            for method in read_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(resp, [b'Some Content'])

            for method in write_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(ro_resp, resp)

    def test_account_read_only_on(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={'sysmeta': {'read-only': 'true'}}):
            for method in read_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(resp, [b'Some Content'])

            for method in write_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(ro_resp, resp)

    def test_account_read_only_off(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={'sysmeta': {'read-only': 'false'}}):
            for method in read_methods + write_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_on_account_off(self):
        conf = {
            'read_only': 'true',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={'sysmeta': {'read-only': 'false'}}):
            for method in read_methods + write_methods:
                req = Request.blank('/v1/a')
                req.method = method
                resp = ro(req.environ, start_response)
                self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_on_allow_deletes(self):
        conf = {
            'read_only': 'true',
            'allow_deletes': 'true',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={}):
            req = Request.blank('/v1/a')
            req.method = "DELETE"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

    def test_account_read_only_on_allow_deletes(self):
        conf = {
            'allow_deletes': 'true',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch('swift.common.middleware.read_only.get_info',
                        return_value={'sysmeta': {'read-only': 'on'}}):
            req = Request.blank('/v1/a')
            req.method = "DELETE"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_on_destination_account_off_on_copy(self):
        conf = {
            'read_only': 'true',
        }

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        def get_fake_read_only(*args, **kwargs):
            if 'b' in args:
                return {'sysmeta': {'read-only': 'false'}}
            return {}

        with mock.patch('swift.common.middleware.read_only.get_info',
                        get_fake_read_only):
            headers = {'Destination-Account': 'b'}
            req = Request.blank('/v1/a', headers=headers)
            req.method = "COPY"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_off_destination_account_on_on_copy(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        def get_fake_read_only(*args, **kwargs):
            if 'b' in args:
                return {'sysmeta': {'read-only': 'true'}}
            return {}

        with mock.patch('swift.common.middleware.read_only.get_info',
                        get_fake_read_only):
            headers = {'Destination-Account': 'b'}
            req = Request.blank('/v1/a', headers=headers)
            req.method = "COPY"
            resp = ro(req.environ, start_response)
            self.assertEqual(ro_resp, resp)

    def test_global_read_only_off_src_acct_on_dest_acct_off_on_copy(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        def fake_account_read_only(self, req, account):
            if account == 'a':
                return 'on'
            return ''

        with mock.patch(
                'swift.common.middleware.read_only.ReadOnlyMiddleware.' +
                'account_read_only',
                fake_account_read_only):
            headers = {'Destination-Account': 'b'}
            req = Request.blank('/v1/a', headers=headers)
            req.method = "COPY"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

    def test_global_read_only_off_src_acct_on_dest_acct_on_on_copy(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        with mock.patch(
                'swift.common.middleware.read_only.ReadOnlyMiddleware.' +
                'account_read_only',
                return_value='true'):
            headers = {'Destination-Account': 'b'}
            req = Request.blank('/v1/a', headers=headers)
            req.method = "COPY"
            resp = ro(req.environ, start_response)
            self.assertEqual(ro_resp, resp)

    def test_global_read_only_non_swift_path(self):
        conf = {}

        ro = read_only.filter_factory(conf)(FakeApp())
        ro.logger = debug_logger()

        def fake_account_read_only(self, req, account):
            return 'on'

        with mock.patch(
                'swift.common.middleware.read_only.ReadOnlyMiddleware.' +
                'account_read_only',
                fake_account_read_only):
            req = Request.blank('/auth/v3.14')
            req.method = "POST"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

            req = Request.blank('/v1')
            req.method = "PUT"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])

            req = Request.blank('/v1.0/')
            req.method = "DELETE"
            resp = ro(req.environ, start_response)
            self.assertEqual(resp, [b'Some Content'])


if __name__ == '__main__':
    unittest.main()
